from polyaxon.k8s.custom_resources.kubeflow.mpi_job import get_mpi_job_custom_resource
from polyaxon.k8s.custom_resources.kubeflow.mx_job import get_mx_job_custom_resource
from polyaxon.k8s.custom_resources.kubeflow.paddle_job import (
    get_paddle_job_custom_resource,
)
from polyaxon.k8s.custom_resources.kubeflow.pytorch_job import (
    get_pytorch_job_custom_resource,
)
from polyaxon.k8s.custom_resources.kubeflow.tf_job import get_tf_job_custom_resource
from polyaxon.k8s.custom_resources.kubeflow.xgb_job import get_xgb_job_custom_resource



from typing import Optional

from pydantic import StrictStr

from polyaxon.lifecycle import V1StageCondition
from polyaxon.schemas.base import BaseResponseModel


class V1EntityStageBodyRequest(BaseResponseModel):
    owner: Optional[StrictStr]
    entity: Optional[StrictStr]
    kind: Optional[StrictStr]
    name: Optional[StrictStr]
    condition: Optional[V1StageCondition]



from typing import Dict

from polyaxon.containers.names import MAIN_JOB_CONTAINER
from polyaxon.polyflow import V1RunKind


class JobMixin:
    K8S_ANNOTATIONS_KIND = V1RunKind.JOB
    MAIN_CONTAINER_ID = MAIN_JOB_CONTAINER


class NotifierMixin:
    K8S_ANNOTATIONS_KIND = V1RunKind.NOTIFIER
    MAIN_CONTAINER_ID = MAIN_JOB_CONTAINER


class CleanerMixin:
    K8S_ANNOTATIONS_KIND = V1RunKind.CLEANER
    MAIN_CONTAINER_ID = MAIN_JOB_CONTAINER


class TunerMixin:
    K8S_ANNOTATIONS_KIND = V1RunKind.TUNER
    MAIN_CONTAINER_ID = MAIN_JOB_CONTAINER


class ServiceMixin:
    K8S_ANNOTATIONS_KIND = V1RunKind.SERVICE
    MAIN_CONTAINER_ID = MAIN_JOB_CONTAINER


MIXIN_MAPPING: Dict = {
    V1RunKind.JOB: JobMixin,
    V1RunKind.NOTIFIER: NotifierMixin,
    V1RunKind.CLEANER: CleanerMixin,
    V1RunKind.TUNER: TunerMixin,
    V1RunKind.SERVICE: ServiceMixin,
}



from typing import List, Optional, Union

from clipped.types.ref_or_obj import RefField
from clipped.types.uuids import UUIDStr
from clipped.utils.lists import to_list
from pydantic import Field, StrictInt, StrictStr, validator

from polyaxon.schemas.types.base import BaseTypeConfig


class V1TensorboardType(BaseTypeConfig):
    """Tensorboard type.

    This type allows to initialize Tensorboard logs foe one or multiple operations.

    Args:
        port: int
        uuids: List[str]
        use_names: bool, optional
        path_prefix: str, optional
        plugins: List[str]

    ### YAML usage

    ### Usage in IO and params definition

    The inputs definition

    ```yaml
    >>> inputs:
    >>>   - name: tensorboard_content
    >>>     type: tensorboard
    ```

    The params usage

    ```yaml
    >>> params:
    >>>   tensorboard_content:
    >>>     value:
    >>>       port: 6006
    >>>       uuids: "d1410a914d18457589b91926d8c23db4,56f1a7f20f1d4f7f9e1a108b3c6b6031"
    >>>       useNames: true
    ```

    ### Usage in initializers

    ```yaml
     ```yaml
    >>> version:  1.1
    >>> kind: component
    >>> run:
    >>>   kind: service
    >>>   init:
    >>>   - tensorboard:
    >>>       uuids: "{{uuids}}"
    >>>       port: "{{globals.ports[0]}}"
    >>>       pathPrefix: "{{globals.base_url}}"
    >>>       useNames: true
    >>>       plugins: "tensorboard-plugin-profile"
    ```

    ### Python usage

    ### Usage in IO and params definition

    The inputs definition

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.polyflow import V1IO
    >>> inputs = [
    >>>     V1IO(
    >>>         name="tensorboard_content",
    >>>         type=types.TENSORBOARD,
    >>>     ),
    >>> ]
    ```

    The params usage

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.polyflow import V1Param
    >>>
    >>> params = {
    >>>     "test1": V1Param(
    >>>         value=types.V1TensorboardType(
    >>>             port=6006,
    >>>             uuids="d1410a914d18457589b91926d8c23db4,56f1a7f20f1d4f7f9e1a108b3c6b6031",
    >>>             use_names=True,
    >>>         )
    >>>     ),
    >>> }
    ```

    ### Usage in initializers

    ```python
    >>> from polyaxon.polyflow import V1Component, V1Init, V1Job
    >>> from polyaxon.schemas.types import V1FileType
    >>> from polyaxon.k8s import k8s_schemas
    >>> component = V1Component(
    >>>     run=V1Job(
    >>>        init=[
    >>>             V1Init(
    >>>                 file=V1TensorboardType(
    >>>                     uuids="{{uuids}}",
    >>>                     port="{{globals.ports[0]}}",
    >>>                     path_prefix="{{globals.base_url}}",
    >>>                     use_names=True,
    >>>                     plugins="tensorboard-plugin-profile",
    >>>                 )
    >>>             ),
    >>>        ],
    >>>        container=k8s_schemas.V1Container(...)
    >>>     )
    >>> )
    ```

    ### Fields
      * port: port to expose the tensorboard service.
      * uuids: comma separated list of operation's uuids to load the tensorboard logs from.
      * useNames: an optional flag to initialize the paths under the operation's names.
      * pathPrefix: an optional path prefix to use for exposing the service.
      * plugins: an optional comma separated list of plugins to install before starting the tensorboard service.
    """

    _IDENTIFIER = "tensorboard"

    port: Optional[Union[StrictInt, RefField]]
    uuids: Optional[Union[List[UUIDStr], RefField]]
    use_names: Optional[Union[bool, RefField]] = Field(alias="useNames")
    path_prefix: Optional[StrictStr] = Field(alias="pathPrefix")
    plugins: Optional[Union[List[StrictStr], RefField]]

    @validator("uuids", "plugins", pre=True)
    def validate_str_list(cls, v, field):
        if isinstance(v, str):
            return to_list(v, check_str=True)
        return v



from typing import List, Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel
from polyaxon.schemas.responses.v1_search import V1Search


class V1ListSearchesResponse(BaseResponseModel):
    count: Optional[int]
    results: Optional[List[V1Search]]
    previous: Optional[StrictStr]
    next: Optional[StrictStr]



def get_container_statuses_by_name(statuses):
    return {
        container_status["name"]: {
            "ready": container_status["ready"],
            "state": container_status["state"],
        }
        for container_status in statuses
    }


def get_container_status(statuses, container_ids):
    job_container_status = None
    for container_id in container_ids:
        job_container_status = statuses.get(container_id)
        if job_container_status:
            break
    return job_container_status



import copy

from typing import Dict, List

from polyaxon.polyaxonfile.specs.libs.parser import PolyaxonfileParser
from polyaxon.polyflow import V1CompiledOperation, V1Operation, V1Param


def get_op_from_schedule(
    content: str,
    compiled_operation: V1CompiledOperation,
) -> V1Operation:
    op_spec = V1Operation.read(content)
    op_spec.conditions = None
    op_spec.schedule = None
    op_spec.events = None
    op_spec.dependencies = None
    op_spec.trigger = None
    op_spec.build = None
    op_spec.skip_on_upstream_skip = None
    op_spec.cache = compiled_operation.cache
    op_spec.queue = compiled_operation.queue
    op_spec.component.inputs = compiled_operation.inputs
    op_spec.component.outputs = compiled_operation.outputs
    op_spec.component.run = compiled_operation.run
    return op_spec


def get_ops_from_suggestions(
    content: str,
    compiled_operation: V1CompiledOperation,
    suggestions: List[Dict],
) -> List[V1Operation]:
    io_keys = compiled_operation.get_io_names()

    def has_param(k: str):
        if not compiled_operation.matrix:
            return None
        return k not in io_keys

    op_content = V1Operation.read(content)  # TODO: Use construct
    for suggestion in suggestions:
        params = {
            k: V1Param(
                value=PolyaxonfileParser.parse_expression(v, {}),
                context_only=has_param(k),
            )
            for (k, v) in suggestion.items()
        }
        op_spec = copy.deepcopy(op_content)
        op_spec.matrix = None
        op_spec.conditions = None
        op_spec.schedule = None
        op_spec.events = None
        op_spec.dependencies = None
        op_spec.trigger = None
        op_spec.build = None
        op_spec.is_approved = None
        op_spec.skip_on_upstream_skip = None
        op_spec.cache = compiled_operation.cache
        op_spec.queue = compiled_operation.queue
        op_spec.params = params
        op_spec.component.inputs = compiled_operation.inputs
        op_spec.component.outputs = compiled_operation.outputs
        op_spec.component.run = compiled_operation.run
        yield op_spec



from polyaxon.k8s.converter.converters.job import JobConverter
from polyaxon.k8s.converter.mixins import CleanerMixin, NotifierMixin, TunerMixin
from polyaxon.utils.fqn_utils import get_cleaner_instance, get_cleaner_resource_name


class NotifierConverter(NotifierMixin, JobConverter):
    pass


class TunerConverter(TunerMixin, JobConverter):
    pass


class CleanerConverter(CleanerMixin, JobConverter):
    def get_instance(self):
        return get_cleaner_instance(
            owner=self.owner_name, project=self.project_name, run_uuid=self.run_uuid
        )

    def get_resource_name(self):
        return get_cleaner_resource_name(self.run_uuid)



from polyaxon.compiler.lineage.artifacts_collector import collect_lineage_artifacts_path
from polyaxon.compiler.lineage.collector import resolve_artifacts_lineage
from polyaxon.compiler.lineage.io_collector import collect_io_artifacts



from datetime import timedelta

from clipped.utils.tz import now

from polyaxon.polyflow import (
    V1HpRange,
    V1HpUniform,
    V1MatrixKind,
    V1Optimization,
    V1RunKind,
)


def get_fxt_schedule_with_inputs_outputs(start_at=None, delta=7):
    start_at = start_at or now()
    start_at = start_at.replace(hour=0, minute=0, second=0, microsecond=0)
    return {
        "version": 1.1,
        "kind": "operation",
        "name": "run",
        "tags": ["key1", "value1"],
        "params": {"image": {"value": "test"}, "lr": {"value": 0.001}},
        "schedule": {
            "kind": "cron",
            "cron": "0 0 * * *",
            "startAt": start_at,
            "endAt": start_at + timedelta(days=delta),
            "dependsOnPast": True,
        },
        "matrix": {
            "kind": V1MatrixKind.BAYES,
            "numInitialRuns": 5,
            "maxIterations": 5,
            "metric": {"name": "loss", "optimization": V1Optimization.MINIMIZE},
            "params": {
                "param1": {"kind": "choice", "value": ["test1", "test2"]},
                "param2": V1HpRange(value=[1, 2, 1]).to_dict(),
                "param3": V1HpUniform(value=[0, 0.9]).to_dict(),
            },
        },
        "component": {
            "name": "experiment-template",
            "description": "experiment to predict something",
            "tags": ["key", "value"],
            "inputs": [
                {"name": "lr", "type": "float", "value": 0.1, "isOptional": True},
                {"name": "image", "type": "str"},
                {"name": "param1", "type": "str"},
                {"name": "param2", "type": "int"},
                {"name": "param3", "type": "float"},
            ],
            "outputs": [
                {"name": "result1", "type": "str"},
                {
                    "name": "result2",
                    "type": "str",
                    "isOptional": True,
                    "value": "{{ image }}",
                },
            ],
            "termination": {"maxRetries": 2},
            "run": {
                "kind": V1RunKind.JOB,
                "environment": {
                    "nodeSelector": {"polyaxon": "experiments"},
                    "serviceAccountName": "service",
                    "imagePullSecrets": ["secret1", "secret2"],
                },
                "container": {
                    "image": "{{ image }}",
                    "command": ["python3", "main.py"],
                    "args": "--lr={{ lr }}",
                    "resources": {"requests": {"cpu": 1}},
                },
            },
        },
    }



from clipped.config.reader import ConfigReader as _ConfigReader

from polyaxon.config.parser import ConfigParser
from polyaxon.config.spec import ConfigSpec


class ConfigReader(_ConfigReader):
    _CONFIG_SPEC = ConfigSpec
    _CONFIG_PARSER = ConfigParser



from typing import Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1RunReferenceCatalog(BaseResponseModel):
    owner: Optional[StrictStr]
    project: Optional[StrictStr]
    name: Optional[StrictStr]
    state: Optional[StrictStr]



from typing import Union
from typing_extensions import Annotated

from pydantic import Field

from polyaxon.polyflow.component.component import V1Component
from polyaxon.polyflow.references import V1DagRef, V1HubRef, V1PathRef, V1UrlRef

V1ComponentReference = Annotated[
    Union[V1Component, V1DagRef, V1HubRef, V1PathRef, V1UrlRef],
    Field(discriminator="kind"),
]



from typing import Dict, Optional

from kubernetes import client as k8s_client
from kubernetes.client import Configuration

from polyaxon import settings
from polyaxon.k8s.converter.converters import CONVERTERS
from polyaxon.k8s.converter.mixins import MIXIN_MAPPING
from polyaxon.runner.executor import BaseExecutor as _BaseExecutor
from polyaxon.runner.kinds import RunnerKind


class BaseExecutor(_BaseExecutor):
    MIXIN_MAPPING = MIXIN_MAPPING
    CONVERTERS = CONVERTERS
    RUNNER_KIND = RunnerKind.K8S

    def __init__(
        self,
        namespace: Optional[str] = None,
        k8s_config: Configuration = None,
        in_cluster: Optional[bool] = None,
    ):
        super().__init__()
        if in_cluster is None:
            in_cluster = settings.CLIENT_CONFIG.in_cluster

        if not namespace:
            namespace = settings.CLIENT_CONFIG.namespace

        self.namespace = namespace
        self.in_cluster = in_cluster
        self.k8s_config = k8s_config

    @classmethod
    def convert(
        cls,
        owner_name: str,
        project_name: str,
        run_name: str,
        run_uuid: str,
        content: str,
        default_auth: bool,
        agent_content: Optional[str] = None,
    ) -> Dict:
        resource = super(BaseExecutor, cls).convert(
            owner_name=owner_name,
            project_name=project_name,
            run_name=run_name,
            run_uuid=run_uuid,
            content=content,
            default_auth=default_auth,
        )
        api = k8s_client.ApiClient()
        return api.sanitize_for_serialization(resource)



import datetime

from typing import Any, Dict, List, Optional, Union

from clipped.types.uuids import UUIDStr
from pydantic import StrictStr

from polyaxon.lifecycle import ManagedBy, V1StatusCondition, V1Statuses
from polyaxon.polyflow.matrix.kinds import V1MatrixKind
from polyaxon.polyflow.run.kinds import V1RunKind
from polyaxon.polyflow.run.resources import V1RunResources
from polyaxon.polyflow.schedules.kinds import V1ScheduleKind
from polyaxon.schemas import V1RunPending
from polyaxon.schemas.base import BaseResponseModel
from polyaxon.schemas.responses.v1_cloning import V1Cloning
from polyaxon.schemas.responses.v1_pipeline import V1Pipeline
from polyaxon.schemas.responses.v1_run_settings import V1RunSettings


class V1Run(BaseResponseModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    description: Optional[StrictStr]
    tags: Optional[List[StrictStr]]
    user: Optional[StrictStr]
    owner: Optional[StrictStr]
    project: Optional[StrictStr]
    schedule_at: Optional[datetime.datetime]
    created_at: Optional[datetime.datetime]
    updated_at: Optional[datetime.datetime]
    started_at: Optional[datetime.datetime]
    finished_at: Optional[datetime.datetime]
    wait_time: Optional[int]
    duration: Optional[int]
    managed_by: Optional[ManagedBy]
    is_managed: Optional[bool]
    is_approved: Optional[bool]
    pending: Optional[V1RunPending]
    content: Optional[StrictStr]
    raw_content: Optional[StrictStr]
    status: Optional[V1Statuses]
    bookmarked: Optional[bool]
    live_state: Optional[int]
    readme: Optional[StrictStr]
    meta_info: Optional[Dict[str, Any]]
    kind: Optional[V1RunKind]
    runtime: Optional[Union[V1RunKind, V1MatrixKind, V1ScheduleKind]]
    inputs: Optional[Dict[str, Any]]
    outputs: Optional[Dict[str, Any]]
    original: Optional[V1Cloning]
    pipeline: Optional[V1Pipeline]
    status_conditions: Optional[List[V1StatusCondition]]
    role: Optional[StrictStr]
    settings: Optional[V1RunSettings]
    resources: Optional[V1RunResources]
    graph: Optional[Dict[str, Any]]
    merge: Optional[bool]



import datetime

from typing import List, Optional

from clipped.types.uuids import UUIDStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel
from polyaxon.schemas.responses.v1_team_settings import V1TeamSettings


class V1Team(BaseResponseModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    projects: Optional[List[StrictStr]]
    component_hubs: Optional[List[StrictStr]]
    model_registries: Optional[List[StrictStr]]
    settings: Optional[V1TeamSettings]
    created_at: Optional[datetime.datetime]
    updated_at: Optional[datetime.datetime]



from typing import Any, Dict, List, Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1ListBookmarksResponse(BaseResponseModel):
    count: Optional[int]
    results: Optional[List[Dict[str, Any]]]
    previous: Optional[StrictStr]
    next: Optional[StrictStr]



from typing import Optional

from polyaxon.schemas.api.version import V1Version
from polyaxon.schemas.base import BaseSchemaModel


class V1Compatibility(BaseSchemaModel):
    cli: Optional[V1Version]
    platform: Optional[V1Version]
    agent: Optional[V1Version]
    ui: Optional[V1Version]



from polyaxon.schemas.base import BaseSchemaModel


class BaseSearchConfig(BaseSchemaModel):
    _USE_DISCRIMINATOR = True

    def create_iteration(self, **kwargs) -> int:
        return 0

    def should_reschedule(self, **kwargs) -> bool:
        return False



from clipped.utils.enums import PEnum


class V1MatrixKind(str, PEnum):
    RANDOM = "random"
    GRID = "grid"
    HYPERBAND = "hyperband"
    BAYES = "bayes"
    HYPEROPT = "hyperopt"
    ITERATIVE = "iterative"
    MAPPING = "mapping"

    @classmethod
    def iteration_values(cls):
        return {
            cls.HYPERBAND,
            cls.BAYES,
            cls.HYPEROPT,
            cls.ITERATIVE,
        }


class V1HPKind(str, PEnum):
    CHOICE = "choice"
    PCHOICE = "pchoice"
    RANGE = "range"
    DATERANGE = "daterange"
    DATETIMERANGE = "datetimerange"
    LINSPACE = "linspace"
    LOGSPACE = "logspace"
    GEOMSPACE = "geomspace"
    UNIFORM = "uniform"
    QUNIFORM = "quniform"
    LOGUNIFORM = "loguniform"
    QLOGUNIFORM = "qloguniform"
    NORMAL = "normal"
    QNORMAL = "qnormal"
    LOGNORMAL = "lognormal"
    QLOGNORMAL = "qlognormal"



import datetime

from typing import Optional

from clipped.types.email import EmailStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1OrganizationMember(BaseResponseModel):
    user: Optional[StrictStr]
    user_email: Optional[EmailStr]
    role: Optional[StrictStr]
    kind: Optional[StrictStr]
    created_at: Optional[datetime.datetime]
    updated_at: Optional[datetime.datetime]



from typing import List, Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseSchemaModel


class V1Installation(BaseSchemaModel):
    key: Optional[StrictStr]
    version: Optional[StrictStr]
    dist: Optional[StrictStr]
    host: Optional[StrictStr]
    hmac: Optional[StrictStr]
    auth: Optional[List[StrictStr]]



from typing import Dict, List, Optional

from clipped.types.numbers import StrictIntOrFloat
from clipped.types.uuids import UUIDStr
from pydantic import StrictInt, StrictStr

from polyaxon.schemas.base import BaseSchemaModel


class ContainerGPUResourcesConfig(BaseSchemaModel):
    _IDENTIFIER = "ContainerGPUResources"
    _MEM_SIZE_ATTRIBUTES = ["memory_free", "memory_used", "memory_total"]

    index: StrictInt
    uuid: UUIDStr
    name: StrictStr
    minor: StrictInt
    bus_id: StrictStr
    serial: StrictStr
    temperature_gpu: StrictIntOrFloat
    utilization_gpu: StrictIntOrFloat
    power_draw: StrictIntOrFloat
    power_limit: StrictIntOrFloat
    memory_free: StrictIntOrFloat
    memory_used: StrictIntOrFloat
    memory_total: StrictIntOrFloat
    memory_utilization: StrictIntOrFloat
    processes: Optional[List[Dict]]


class ContainerResourcesConfig(BaseSchemaModel):
    _IDENTIFIER = "ContainerResources"
    _PERCENT_ATTRIBUTES = ["cpu_percentage"]
    _MEM_SIZE_ATTRIBUTES = ["memory_used", "memory_limit"]

    job_uuid: UUIDStr
    experiment_uuid: UUIDStr
    job_name: StrictStr
    container_id: StrictStr
    n_cpus: StrictIntOrFloat
    cpu_percentage: StrictIntOrFloat
    percpu_percentage: Optional[List[float]]
    memory_used: StrictIntOrFloat
    memory_limit: StrictIntOrFloat
    gpu_resources: Optional[List[ContainerGPUResourcesConfig]]



from typing import List, Optional

from polyaxon.schemas.base import BaseResponseModel
from polyaxon.schemas.responses.v1_settings_catalog import V1SettingsCatalog


class V1TeamSettings(BaseResponseModel):
    projects: Optional[List[V1SettingsCatalog]]
    hubs: Optional[List[V1SettingsCatalog]]



import logging

from typing import Dict, Optional

logger = logging.getLogger("polyaxon.cli")


def set_raven_client(options: Optional[Dict] = None) -> bool:
    from polyaxon import pkg, settings
    from polyaxon.services.values import PolyaxonServices

    cli_config = settings.CLI_CONFIG
    options = options or {}
    environment = options.get("environment")
    dsn = options.get("dsn")
    sample_rate = options.get("sample_rate", 0)
    if cli_config and cli_config.log_handler and cli_config.log_handler.decoded_dsn:
        dsn = dsn or cli_config.log_handler.decoded_dsn
        environment = environment or cli_config.log_handler.environment

    if dsn:
        import sentry_sdk

        sentry_sdk.init(
            dsn=dsn,
            release=pkg.VERSION,
            environment=environment,
            server_name=PolyaxonServices.get_service_name(),
            sample_rate=sample_rate,
        )
        return True

    return False



import functools

from typing import Callable, Optional

from urllib3.exceptions import HTTPError

from polyaxon import settings
from polyaxon.client import PolyaxonClient
from polyaxon.client.decorators.errors import handle_client_error
from polyaxon.logger import logger
from polyaxon.sdk.exceptions import ApiException


def client_handler(
    check_no_op: bool = True,
    check_offline: bool = False,
    can_log_events: bool = False,
    can_log_outputs: bool = False,
) -> Callable:
    """
    The `ClientHandlerDecorator` is a decorator to handle several checks in PolyaxonClient.

     * check_offline: to ignore any decorated function when POLYAXON_IS_OFFLINE env var is found.
     * check_no_op: to ignore any decorated function when NO_OP env var is found.
     * handle_openapi_exceptions: to handle exception of OpenApi and generate better
        debugging outputs.
     * can_log_events: to check if there's an event logger instance on the object.
     * can_log_outputs: to check if there's an outputs path set on the run.
     * openapi_extra_context: to augment openapi errors.

    usage example with class method:
        @client_handler(check_no_op=True)
        def my_func(self, *args, **kwargs):
            ...
            return ...

    usage example with a function:
        @client_handler(check_no_op=True)
        def my_func(arg1, arg2):
            ...
            return ...
    """

    def _check_global_or_inline_config(args, config_key) -> Optional[bool]:
        self_arg = args[0] if args else None
        config_value = getattr(self_arg, f"_{config_key}", None)
        return get_global_or_inline_config(
            config_key=config_key,
            config_value=config_value,
            client=getattr(self_arg, "_client", None),
        )

    def client_handler_wrapper(f: Callable) -> Callable:
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if check_no_op and _check_global_or_inline_config(args, "no_op"):
                logger.debug("Using NO_OP mode")
                return None
            if check_offline and _check_global_or_inline_config(args, "is_offline"):
                logger.debug("Using IS_OFFLINE mode")
                return None
            manual_exceptions_handling = False
            if args:
                self_arg = args[0]

                manual_exceptions_handling = getattr(
                    self_arg, "_manual_exceptions_handling", False
                )

                if can_log_events and (
                    not hasattr(self_arg, "_event_logger")
                    or self_arg._event_logger is None  # pylint:disable=protected-access
                ):
                    logger.warning(
                        "You should set an event logger before calling: {}".format(
                            f.__name__
                        )
                    )

                if can_log_outputs and (
                    not hasattr(self_arg, "_outputs_path")
                    or self_arg._outputs_path is None  # pylint:disable=protected-access
                ):
                    logger.warning(
                        "You should set an an outputs path before calling: {}".format(
                            f.__name__
                        )
                    )

            try:
                return f(*args, **kwargs)
            except (ApiException, HTTPError) as e:
                message = (
                    "\nAPI Client failed at the function `%(name)s` in file "
                    "`%(filename)s` line number `%(line)s`."
                    "\nClient config:\n%(config)s\n"
                    % {
                        "name": f.__name__,
                        "filename": f.__code__.co_filename,
                        "line": f.__code__.co_firstlineno + 1,
                        "config": settings.CLIENT_CONFIG.to_dict(),
                    }
                )
                handle_client_error(
                    e=None if manual_exceptions_handling else e, message=message
                )
                raise e

        return wrapper

    return client_handler_wrapper


def get_global_or_inline_config(
    config_key: str,
    config_value: Optional[bool] = None,
    client: Optional[PolyaxonClient] = None,
) -> Optional[bool]:
    if config_value is not None:
        return config_value
    if client and client.config and getattr(client.config, config_key) is not None:
        return getattr(client.config, config_key)
    return getattr(settings.CLIENT_CONFIG, config_key, None)



from typing import List, Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel
from polyaxon.schemas.responses.v1_project_user_access import V1ProjectUserAccess


class V1ProjectSettings(BaseResponseModel):
    connections: Optional[List[StrictStr]]
    preset: Optional[StrictStr]
    presets: Optional[List[StrictStr]]
    queue: Optional[StrictStr]
    queues: Optional[List[StrictStr]]
    agents: Optional[List[StrictStr]]
    user_accesses: Optional[List[V1ProjectUserAccess]]
    teams: Optional[List[StrictStr]]
    projects: Optional[List[StrictStr]]



from clipped.utils.enums import PEnum

from polyaxon.lifecycle import V1Statuses


class V1TriggerPolicy(str, PEnum):
    ALL_SUCCEEDED = "all_succeeded"
    ALL_FAILED = "all_failed"
    ALL_DONE = "all_done"
    ONE_SUCCEEDED = "one_succeeded"
    ONE_FAILED = "one_failed"
    ONE_DONE = "one_done"

    @classmethod
    def trigger_statuses_mapping(cls):
        return {
            cls.ALL_SUCCEEDED: V1Statuses.SUCCEEDED,
            cls.ALL_FAILED: V1Statuses.FAILED,
            cls.ALL_DONE: V1Statuses.DONE,
            cls.ONE_SUCCEEDED: V1Statuses.SUCCEEDED,
            cls.ONE_FAILED: V1Statuses.FAILED,
            cls.ONE_DONE: V1Statuses.DONE,
        }



# This code is based on logic from
# http://sqizit.bartletts.id.au/2011/02/14/pseudo-terminals-in-python/
# Licensed under the MIT license:
# Copyright (c) 2011 Joshua D. Bartlett
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import fcntl
import os
import pty
import select
import signal
import struct
import termios
import tty

from clipped.utils.json import orjson_dumps
from polyaxon.client.transport import ws_client


class PseudoTerminal:
    """Wraps the pseudo-TTY (PTY) allocated to a container.

    The PTY is managed via the current process' TTY until it is closed.
    """

    START_ALTERNATE_MODE = set("\x1b[?{0}h".format(i) for i in ("1049", "47", "1047"))
    END_ALTERNATE_MODE = set("\x1b[?{0}l".format(i) for i in ("1049", "47", "1047"))
    ALTERNATE_MODE_FLAGS = tuple(START_ALTERNATE_MODE) + tuple(END_ALTERNATE_MODE)

    def __init__(self, client_shell=None):
        self.client_shell = client_shell
        self.master_fd = None

    def start(self, argv=None):
        """
        Create a spawned process.
        Based on the code for pty.spawn().
        """
        if not argv:
            argv = [os.environ["SHELL"]]

        pid, master_fd = pty.fork()
        self.master_fd = master_fd
        if pid == pty.CHILD:
            os.execlp(argv[0], *argv)

        old_handler = signal.signal(signal.SIGWINCH, self._signal_winch)
        try:
            mode = tty.tcgetattr(pty.STDIN_FILENO)
            tty.setraw(pty.STDIN_FILENO)
            restore = 1
        except tty.error:  # This is the same as termios.error
            restore = 0
        self._init_fd()
        try:
            self._loop()
        except (IOError, OSError):
            if restore:
                tty.tcsetattr(pty.STDIN_FILENO, tty.TCSAFLUSH, mode)

        self.client_shell.close()
        self.client_shell = None
        if self.master_fd:
            os.close(self.master_fd)
            self.master_fd = None
        signal.signal(signal.SIGWINCH, old_handler)

    def _init_fd(self):
        """
        Called once when the pty is first set up.
        """
        self._set_pty_size()

    def _signal_winch(self, signum, frame):
        """
        Signal handler for SIGWINCH - window size has changed.
        """
        self._set_pty_size()

    def _set_pty_size(self):
        """
        Sets the window size of the child pty based on the window size of
        our own controlling terminal.
        """
        packed = fcntl.ioctl(
            pty.STDOUT_FILENO, termios.TIOCGWINSZ, struct.pack("HHHH", 0, 0, 0, 0)
        )
        rows, cols, h_pixels, v_pixels = struct.unpack("HHHH", packed)
        self.client_shell.write_channel(
            ws_client.RESIZE_CHANNEL, orjson_dumps({"Height": rows, "Width": cols})
        )

    def _loop(self):
        """
        Main select loop. Passes all data to self.master_read() or self.stdin_read().
        """
        assert self.client_shell is not None
        client_shell = self.client_shell
        while 1:
            try:
                rfds, wfds, xfds = select.select(
                    [pty.STDIN_FILENO, client_shell.sock.sock], [], []
                )
            except select.error as e:
                no = e.errno
                if no == errno.EINTR:
                    continue
            if pty.STDIN_FILENO in rfds:
                data = os.read(pty.STDIN_FILENO, 1024)
                self.stdin_read(data)
            if client_shell.sock.sock in rfds:
                # read from client_shell
                if client_shell.peek_stdout():
                    self.master_read(client_shell.read_stdout())
                if client_shell.peek_stderr():
                    self.master_read(client_shell.read_stderr())
                # error occurs
                if client_shell.peek_channel(ws_client.ERROR_CHANNEL):
                    break

    def write_stdout(self, data):
        """
        Writes to stdout as if the child process had written the data.
        """
        os.write(pty.STDOUT_FILENO, data.encode())

    def write_master(self, data):
        """
        Writes to the child process from its controlling terminal.
        """
        assert self.client_shell is not None
        self.client_shell.write_stdin(data)

    def master_read(self, data):
        """
        Called when there is data to be sent from the child process back to the user.
        """
        flag = self.findlast(data, self.ALTERNATE_MODE_FLAGS)
        if flag is not None:
            if flag in self.START_ALTERNATE_MODE:
                # This code is executed when the child process switches the
                #       terminal into alternate mode. The line below
                #       assumes that the user has opened vim, and writes a
                #       message.
                self.write_master("IEntering special mode.\x1b")
            elif flag in self.END_ALTERNATE_MODE:
                # This code is executed when the child process switches the
                #       terminal back out of alternate mode. The line below
                #       assumes that the user has returned to the command
                #       prompt.
                self.write_master('echo "Leaving special mode."\r')
        self.write_stdout(data)

    def stdin_read(self, data):
        """
        Called when there is data to be sent from the user/controlling
        terminal down to the child process.
        """
        self.write_master(data)

    @staticmethod
    def findlast(s, substrs):
        """
        Finds whichever of the given substrings occurs last in the given string
        and returns that substring, or returns None if no such strings occur.
        """
        i = -1
        result = None
        for substr in substrs:
            pos = s.rfind(substr)
            if pos > i:
                i = pos
                result = substr
        return result



# All schemas used by Polyaxon

from kubernetes import client

V1Affinity = client.V1Affinity
V1SecurityContext = client.V1SecurityContext
V1PodDNSConfig = client.V1PodDNSConfig
V1Toleration = client.V1Toleration
V1HostAlias = client.V1HostAlias
V1Container = client.V1Container
V1Container.name = ""
V1EnvVar = client.V1EnvVar
V1VolumeMount = client.V1VolumeMount
V1ContainerPort = client.V1ContainerPort
V1ResourceRequirements = client.V1ResourceRequirements
V1EnvFromSource = client.V1EnvFromSource
V1Volume = client.V1Volume
V1ObjectFieldSelector = client.V1ObjectFieldSelector
V1EnvVarSource = client.V1EnvVarSource
V1ConfigMapKeySelector = client.V1ConfigMapKeySelector
V1SecretKeySelector = client.V1SecretKeySelector
V1PodSpec = client.V1PodSpec
V1ObjectMeta = client.V1ObjectMeta
V1PodTemplateSpec = client.V1PodTemplateSpec
V1HostPathVolumeSource = client.V1HostPathVolumeSource
V1EmptyDirVolumeSource = client.V1EmptyDirVolumeSource
V1PersistentVolumeClaimVolumeSource = client.V1PersistentVolumeClaimVolumeSource
V1SecretVolumeSource = client.V1SecretVolumeSource
V1ConfigMapVolumeSource = client.V1ConfigMapVolumeSource
V1LocalObjectReference = client.V1LocalObjectReference



from typing import List, Optional

from polyaxon.schemas.base import BaseResponseModel


class V1AgentStateResponseAgentState(BaseResponseModel):
    schedules: Optional[List]
    hooks: Optional[List]
    watchdogs: Optional[List]
    tuners: Optional[List]
    queued: Optional[List]
    stopping: Optional[List]
    deleting: Optional[List]
    apply: Optional[List]
    checks: Optional[List]
    full: Optional[bool]



from collections.abc import Mapping
from typing import Dict, List, Union

from clipped.formatting import Printer

from polyaxon.exceptions import PolyaxonfileError
from polyaxon.polyaxonfile.specs.libs.parser import PolyaxonfileParser


def parse_params(params: Union[List[str], Dict], is_cli: bool = True):
    if isinstance(params, Mapping):
        return {
            k: {"value": PolyaxonfileParser.parse_expression(v, {})}
            for k, v in params.items()
        }

    parsed_params = {}
    for param in params:
        index = param.find("=")
        if index == -1:
            message = (
                "Invalid format for -P parameter: '%s'. Use -P name=value." % param
            )
            if is_cli:
                Printer.error(message, sys_exit=True)
            else:
                raise PolyaxonfileError(message)
        name = param[:index]
        value = param[index + 1 :]
        if name in parsed_params:
            message = "Repeated parameter: '%s'" % name
            if is_cli:
                Printer.error(message, sys_exit=True)
            else:
                raise PolyaxonfileError(message)
        parsed_params[name] = {"value": value}

    return parsed_params


def parse_hparams(hparams: Union[List[str], Dict], is_cli: bool = True):
    if isinstance(hparams, Mapping):
        return hparams

    parsed_params = {}
    for param in hparams:
        index = param.find("=")
        if index == -1:
            message = "Invalid format for -HP parameter: `{}`. Use -HP name='kind:value'.".format(
                param
            )
            if is_cli:
                Printer.error(message, sys_exit=True)
            else:
                raise PolyaxonfileError(message)
        name = param[:index]
        value = param[index + 1 :]
        if name in parsed_params:
            message = "Repeated parameter: `{}`".format(name)
            if is_cli:
                Printer.error(message, sys_exit=True)
            else:
                raise PolyaxonfileError(message)
        values = value.split(":")
        if len(values) < 2:
            message = (
                "Hyper-parameter: `{}` received an incorrect value format `{}`".format(
                    name, value
                )
            )
            if is_cli:
                Printer.error(message, sys_exit=True)
            else:
                raise PolyaxonfileError(message)
        kind = values[0]
        try:
            from polyaxon.config.spec import ConfigSpec

            value = ConfigSpec.read_from_stream(":".join(values[1:]))
        except Exception as e:
            message = (
                "Hyper-parameter: `{}` with kind `{}` received an incorrect value `{}`. "
                "Parsing error: {}".format(name, kind, value, e)
            )
            if is_cli:
                Printer.error(message, sys_exit=True)
            else:
                raise PolyaxonfileError(message)
        parsed_params[name] = {"kind": kind, "value": value}

    return parsed_params



from typing import Optional

from pydantic import Field, StrictStr

from polyaxon.schemas.base import BaseSchemaModel


class SSLConfig(BaseSchemaModel):
    enabled: Optional[bool]
    secret_name: Optional[StrictStr] = Field(alias="secretName")
    path: Optional[StrictStr]



import os

from typing import TYPE_CHECKING, Optional

from clipped.formatting import Printer
from clipped.utils.bools import to_bool
from pydantic import ValidationError

from polyaxon.api import LOCALHOST
from polyaxon.env_vars.keys import EV_KEYS_NO_CONFIG, EV_KEYS_SET_AGENT
from polyaxon.managers.client import ClientConfigManager
from polyaxon.managers.home import HomeConfigManager
from polyaxon.managers.user import UserConfigManager
from polyaxon.services.values import PolyaxonServices

if TYPE_CHECKING:
    from polyaxon.deploy.schemas.auth import AccessTokenConfig
    from polyaxon.schemas.api.home import HomeConfig
    from polyaxon.schemas.cli.agent_config import AgentConfig
    from polyaxon.schemas.cli.cli_config import CliConfig
    from polyaxon.schemas.cli.client_config import ClientConfig

MIN_TIMEOUT = 1
LONG_REQUEST_TIMEOUT = 3600
HEALTH_CHECK_INTERVAL = 60
SET_AGENT = to_bool(os.environ.get(EV_KEYS_SET_AGENT, False))

HOME_CONFIG: "HomeConfig" = HomeConfigManager.get_config_from_env()
AUTH_CONFIG: Optional["AccessTokenConfig"] = None
CLIENT_CONFIG: "ClientConfig"
CLI_CONFIG: Optional["CliConfig"] = None
AGENT_CONFIG: Optional["AgentConfig"] = None

PolyaxonServices.set_service_name()


def set_agent_config(config: Optional["AgentConfig"] = None):
    from polyaxon.connections import CONNECTION_CONFIG
    from polyaxon.managers.agent import AgentConfigManager

    global AGENT_CONFIG

    # Patch the config with correct home path if available
    AgentConfigManager.set_config_path(HOME_CONFIG.path)

    AGENT_CONFIG = config or AgentConfigManager.get_config_from_env()

    # Set artifacts store name
    AGENT_CONFIG.set_artifacts_store_name()

    # Always sync the connections catalog to the current agent config
    CONNECTION_CONFIG.set_connections_catalog(AGENT_CONFIG.all_connections)


def set_cli_config():
    from polyaxon.managers.cli import CliConfigManager

    global CLI_CONFIG

    # Patch the config with correct home path if available
    CliConfigManager.set_config_path(HOME_CONFIG.path)

    try:
        CLI_CONFIG = CliConfigManager.get_config_or_default()
    except (TypeError, ValidationError):
        CliConfigManager.purge()
        Printer.warning("Your CLI configuration was purged!")


def set_client_config():
    global CLIENT_CONFIG

    # Patch the config with correct home path if available
    ClientConfigManager.set_config_path(HOME_CONFIG.path)

    try:
        CLIENT_CONFIG = ClientConfigManager.get_config_from_env()
    except (TypeError, ValidationError):
        ClientConfigManager.purge()
        Printer.warning("Your client configuration was purged!")
        CLIENT_CONFIG = ClientConfigManager.get_config_from_env()


def set_auth_config():
    from polyaxon.managers.auth import AuthConfigManager

    global AUTH_CONFIG

    # Patch the config with correct home path if available
    AuthConfigManager.set_config_path(HOME_CONFIG.path)

    try:
        AUTH_CONFIG = AuthConfigManager.get_config_from_env()
    except (TypeError, ValidationError):
        AuthConfigManager.purge()
        Printer.warning("Your auth configuration was purged!")

    # Patch the config with correct home path if available
    UserConfigManager.set_config_path(HOME_CONFIG.path)

    try:
        UserConfigManager.get_config_or_default()
    except (TypeError, ValidationError):
        UserConfigManager.purge()
        Printer.warning("Your user configuration was purged!")


if not to_bool(os.environ.get(EV_KEYS_NO_CONFIG, False)):
    set_auth_config()
    set_client_config()
    if PolyaxonServices.is_agent() or SET_AGENT:
        set_agent_config()
else:
    # Patch the config with correct home path if available
    ClientConfigManager.set_config_path(HOME_CONFIG.path)

    CLIENT_CONFIG = ClientConfigManager.CONFIG(host=LOCALHOST)



from polyaxon.polyflow.references.dag import V1DagRef
from polyaxon.polyflow.references.hub import V1HubRef
from polyaxon.polyflow.references.mixin import RefMixin
from polyaxon.polyflow.references.path import V1PathRef
from polyaxon.polyflow.references.url import V1UrlRef



from typing import List, Optional

from polyaxon.auxiliaries import get_default_cleaner_container
from polyaxon.auxiliaries.cleaner import V1PolyaxonCleaner, get_batch_cleaner_container
from polyaxon.connections import V1Connection
from polyaxon.polyflow import (
    V1CleanerJob,
    V1Component,
    V1Environment,
    V1Operation,
    V1Plugins,
    V1Termination,
)


def get_cleaner_operation(
    connection: V1Connection,
    run_uuid: str,
    run_kind: str,
    environment: Optional[V1Environment] = None,
    cleaner: Optional[V1PolyaxonCleaner] = None,
) -> V1Operation:
    return V1Operation(
        termination=V1Termination(max_retries=1),
        component=V1Component(
            name="cleaner",
            plugins=V1Plugins(
                auth=False,
                collect_logs=False,
                collect_artifacts=False,
                collect_resources=False,
                auto_resume=False,
                sync_statuses=False,
            ),
            run=V1CleanerJob(
                environment=environment,
                connections=[connection.name],
                container=get_default_cleaner_container(
                    connection, run_uuid, run_kind, cleaner
                ),
            ),
        ),
    )


def get_batch_cleaner_operation(
    connection: V1Connection,
    paths: List[str],
    environment: Optional[V1Environment] = None,
    cleaner: Optional[V1PolyaxonCleaner] = None,
) -> V1Operation:
    return V1Operation(
        termination=V1Termination(max_retries=1),
        component=V1Component(
            name="cleaner",
            plugins=V1Plugins(
                auth=False,
                collect_logs=False,
                collect_artifacts=False,
                collect_resources=False,
                auto_resume=False,
                sync_statuses=False,
            ),
            run=V1CleanerJob(
                environment=environment,
                connections=[connection.name],
                container=get_batch_cleaner_container(connection, paths, cleaner),
            ),
        ),
    )



from typing import List, Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseSchemaModel


class V1ArtifactsMount(BaseSchemaModel):
    _IDENTIFIER = "artifacts_mount"

    name: StrictStr
    paths: Optional[List[StrictStr]]



import sys

from clipped.formatting import Printer

from polyaxon.constants.globals import DEFAULT
from polyaxon.exceptions import PolyaxonClientException
from polyaxon.managers.user import UserConfigManager


def get_local_owner(is_cli: bool = False):
    from polyaxon import settings

    owner = None
    if UserConfigManager.is_initialized():
        try:
            user_config = UserConfigManager.get_config()
            owner = user_config.organization if user_config else None
        except TypeError:
            Printer.error(
                "Found an invalid user config or user config cache, "
                "if you are using Polyaxon CLI please run: "
                "`polyaxon config purge --cache-only`",
                sys_exit=True,
            )

    if not owner and (not settings.CLI_CONFIG or settings.CLI_CONFIG.is_community):
        owner = DEFAULT

    if not owner:
        error = "An context owner (user or organization) is required."
        if is_cli:
            Printer.error(error)
            sys.exit(1)
        else:
            raise PolyaxonClientException(error)
    return owner



from typing import List, Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel
from polyaxon.schemas.responses.v1_team_member import V1TeamMember


class V1ListTeamMembersResponse(BaseResponseModel):
    count: Optional[int]
    results: Optional[List[V1TeamMember]]
    previous: Optional[StrictStr]
    next: Optional[StrictStr]



from typing import Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1PasswordChange(BaseResponseModel):
    old_password: Optional[StrictStr]
    new_password1: Optional[StrictStr]
    new_password2: Optional[StrictStr]



from typing import Dict, List, Optional, Union
from typing_extensions import Literal

from clipped.formatting import Printer
from pydantic import Field, StrictInt, StrictStr, root_validator, validator

from polyaxon.auxiliaries import (
    V1PolyaxonCleaner,
    V1PolyaxonInitContainer,
    V1PolyaxonNotifier,
    V1PolyaxonSidecarContainer,
)
from polyaxon.auxiliaries.default_scheduling import V1DefaultScheduling
from polyaxon.connections import V1Connection
from polyaxon.deploy.schemas.auth import AuthConfig
from polyaxon.deploy.schemas.deployment_types import DeploymentCharts, DeploymentTypes
from polyaxon.deploy.schemas.email import EmailConfig
from polyaxon.deploy.schemas.ingress import IngressConfig
from polyaxon.deploy.schemas.intervals import IntervalsConfig
from polyaxon.deploy.schemas.operators import OperatorsConfig
from polyaxon.deploy.schemas.proxy import ProxyConfig
from polyaxon.deploy.schemas.rbac import RBACConfig
from polyaxon.deploy.schemas.root_user import RootUserConfig
from polyaxon.deploy.schemas.security_context import SecurityContextConfig
from polyaxon.deploy.schemas.service import (
    AgentServiceConfig,
    ApiServiceConfig,
    DeploymentService,
    ExternalServicesConfig,
    HooksConfig,
    OperatorServiceConfig,
    PostgresqlConfig,
    RabbitmqConfig,
    RedisConfig,
    WorkerServiceConfig,
)
from polyaxon.deploy.schemas.ssl import SSLConfig
from polyaxon.deploy.schemas.ui import UIConfig
from polyaxon.k8s import k8s_schemas, k8s_validation
from polyaxon.schemas.base import BaseSchemaModel


def validate_connections(artifacts_store, connections):
    connections = connections or []

    connection_names = set()

    if artifacts_store:
        connection_names.add(artifacts_store.name)

    for c in connections:
        if c.name in connection_names:
            raise ValueError(
                "A connection with name `{}` must be unique.".format(c.name)
            )
        connection_names.add(c.name)


def check_postgres(postgresql, external_services):
    postgresql_disabled = postgresql.enabled is False if postgresql else False
    external_postgresql = None
    if external_services:
        external_postgresql = external_services.postgresql

    if postgresql_disabled and not external_postgresql:
        raise ValueError(
            "A postgresql instance is required, "
            "please enable the in-cluster postgresql, "
            "or provide an external instance."
        )


def check_rabbitmq(rabbitmq, external_services, broker):
    rabbitmq_enabled = rabbitmq and rabbitmq.enabled
    external_rabbitmq = None
    rabbitmq_broker = broker != "redis"
    if external_services:
        external_rabbitmq = external_services.rabbitmq

    if rabbitmq_enabled and external_rabbitmq:
        raise ValueError(
            "You can either enable the in-cluster rabbitmq or use an external instance, "
            "not both!"
        )
    rabbitmq_used = rabbitmq_enabled or external_rabbitmq
    if rabbitmq_used and not rabbitmq_broker:
        raise ValueError(
            "rabbitmq is enabled but you are using a different broker backend!"
        )

    return rabbitmq_used


def check_redis(redis, external_services, broker):
    redis_enabled = redis and redis.enabled
    redis_non_broker = redis and redis.non_broker
    external_redis = None
    redis_broker = broker == "redis"
    if external_services:
        external_redis = external_services.redis

    if redis_enabled and external_redis:
        raise ValueError(
            "You can either enable the in-cluster redis or use an external instance, "
            "not both!"
        )

    redis_used = redis_enabled or external_redis
    if redis_used and not redis_broker and not redis_non_broker:
        raise ValueError(
            "redis is enabled but you are using a different broker backend!"
        )

    return redis_used, not redis_non_broker


def broker_is_required(services):
    for s in services:
        if s and s.enabled:
            return True
    return False


def wrong_agent_deployment_keys(**kwargs):
    error_keys = []
    for k, v in kwargs.items():
        if v is not None:
            error_keys.append(k)
    if error_keys:
        raise ValueError(
            "Agent deployment received some keys that are not required.\n"
            "Please remove these config keys from your config file:\n"
            "{}".format(error_keys)
        )


def validate_platform_deployment(
    postgresql,
    redis,
    rabbitmq,
    broker,
    scheduler,
    compiler,
    worker,
    beat,
    external_services,
):
    check_postgres(postgresql, external_services)
    redis_used, redis_is_broker = check_redis(redis, external_services, broker)
    rabbitmq_used = check_rabbitmq(rabbitmq, external_services, broker)
    if rabbitmq_used and redis_used and redis_is_broker:
        raise ValueError(
            "You only need to enable rabbitmq or redis for the broker, "
            "you don't need to deploy both!"
        )
    broker_defined = rabbitmq_used or redis_used
    services = [scheduler, compiler, worker, beat]
    if broker_is_required(services) and not broker_defined:
        raise ValueError(
            "You enabled some services that require a broker, please set redis or rabbitmq!"
        )


def validate_deployment_chart(
    deployment_chart,
    agent,
    environment,
):
    if deployment_chart == DeploymentCharts.AGENT and not agent:
        raise ValueError("Agent deployment requires a valid `agent` key configuration.")

    if deployment_chart == DeploymentCharts.PLATFORM and agent:
        Printer.header("Deployment has agent enabled!")


class DeploymentConfig(BaseSchemaModel):
    _SWAGGER_FIELDS = [
        "tolerations",
        "affinity",
        "celeryTolerations",
        "celeryAffinity",
    ]

    deployment_type: Optional[DeploymentTypes] = Field(alias="deploymentType")
    deployment_chart: Optional[DeploymentCharts] = Field(
        default=DeploymentCharts.PLATFORM, alias="deploymentChart"
    )
    deployment_version: Optional[StrictStr] = Field(alias="deploymentVersion")
    release_name: Optional[StrictStr] = Field(alias="releaseName")
    namespace: Optional[StrictStr]
    rbac: Optional[RBACConfig]
    polyaxon_secret: Optional[StrictStr] = Field(alias="polyaxonSecret")
    internal_token: Optional[StrictStr] = Field(alias="internalToken")
    password_length: Optional[StrictInt] = Field(alias="passwordLength")
    ssl: Optional[SSLConfig]
    encryption_secret: Optional[StrictStr] = Field(alias="encryptionSecret")
    platform_secret: Optional[StrictStr] = Field(alias="platformSecret")
    agent_secret: Optional[StrictStr] = Field(alias="agentSecret")
    timezone: Optional[StrictStr]
    environment: Optional[StrictStr]
    ingress: Optional[IngressConfig]
    user: Optional[RootUserConfig]
    node_selector: Optional[Dict[StrictStr, StrictStr]] = Field(alias="nodeSelector")
    tolerations: Optional[List[Union[k8s_schemas.V1Toleration, Dict]]]
    affinity: Optional[Union[k8s_schemas.V1Affinity, Dict]]
    labels: Optional[Dict[StrictStr, StrictStr]] = Field(alias="labels")
    annotations: Optional[Dict[StrictStr, StrictStr]] = Field(alias="annotations")
    priority_class_name: Optional[StrictStr] = Field(alias="priorityClassName")
    celery_node_selector: Optional[Dict[StrictStr, StrictStr]] = Field(
        alias="celeryNodeSelector"
    )
    celery_tolerations: Optional[List[Union[k8s_schemas.V1Toleration, Dict]]] = Field(
        alias="celeryTolerations"
    )
    celery_affinity: Optional[Union[k8s_schemas.V1Affinity, Dict]] = Field(
        alias="celeryAffinity"
    )
    limit_resources: Optional[bool] = Field(alias="limitResources")
    global_replicas: Optional[StrictInt] = Field(alias="globalReplicas")
    global_concurrency: Optional[StrictInt] = Field(alias="globalConcurrency")
    gateway: Optional[ApiServiceConfig]
    scheduler: Optional[WorkerServiceConfig]
    compiler: Optional[WorkerServiceConfig]
    worker: Optional[WorkerServiceConfig]
    beat: Optional[DeploymentService]
    agent: Optional[AgentServiceConfig]
    operator: Optional[OperatorServiceConfig]
    init: Optional[V1PolyaxonInitContainer]
    sidecar: Optional[V1PolyaxonSidecarContainer]
    notifier: Optional[V1PolyaxonNotifier]
    cleaner: Optional[V1PolyaxonCleaner]
    default_scheduling: Optional[V1DefaultScheduling] = Field(alias="defaultScheduling")
    default_image_pull_secrets: Optional[List[StrictStr]] = Field(
        alias="defaultImagePullSecrets"
    )
    clean_hooks: Optional[HooksConfig] = Field(alias="cleanHooks")
    api_hooks: Optional[HooksConfig] = Field(alias="apiHooks")
    flower: Optional[DeploymentService]
    postgresql: Optional[PostgresqlConfig]
    redis: Optional[RedisConfig]
    rabbitmq: Optional[RabbitmqConfig]
    broker: Optional[Literal["redis", "rabbitmq"]]
    email: Optional[EmailConfig]
    ldap: Optional[Dict]
    metrics: Optional[Dict]
    image_pull_secrets: Optional[List[StrictStr]] = Field(alias="imagePullSecrets")
    host_name: Optional[StrictStr] = Field(alias="hostName")
    allowed_hosts: Optional[List[StrictStr]] = Field(alias="allowedHosts")
    include_host_ips: Optional[bool] = Field(alias="includeHostIps")
    intervals: Optional[IntervalsConfig]
    artifacts_store: Optional[V1Connection] = Field(alias="artifactsStore")
    connections: Optional[List[V1Connection]]
    mount_connections: Optional[List[str]] = Field(alias="mountConnections")
    log_level: Optional[StrictStr] = Field(alias="logLevel")
    security_context: Optional[SecurityContextConfig] = Field(alias="securityContext")
    external_services: Optional[ExternalServicesConfig] = Field(
        alias="externalServices"
    )
    debug_mode: Optional[bool] = Field(alias="debugMode")
    organization_key: Optional[StrictStr] = Field(alias="organizationKey")
    auth: Optional[AuthConfig]
    proxy: Optional[ProxyConfig]
    ui: Optional[UIConfig]
    include_chart_revision: Optional[bool] = Field(alias="includeChartRevision")
    operators: Optional[OperatorsConfig]
    istio: Optional[Dict]
    dns: Optional[Dict]

    @root_validator
    def validate_deployment(cls, values):
        validate_deployment_chart(
            deployment_chart=values.get("deployment_chart"),
            agent=values.get("agent"),
            environment=values.get("environment"),
        )
        validate_platform_deployment(
            postgresql=values.get("postgresql"),
            redis=values.get("redis"),
            rabbitmq=values.get("rabbitmq"),
            broker=values.get("broker"),
            scheduler=values.get("scheduler"),
            compiler=values.get("compiler"),
            worker=values.get("worker"),
            beat=values.get("beat"),
            external_services=values.get("external_services"),
        )
        if values.get("deployment_chart") == DeploymentCharts.AGENT:
            wrong_agent_deployment_keys(
                password_length=values.get("password_length"),
                platform_secret=values.get("platform_secret"),
                encryption_secret=values.get("encryption_secret"),
                user=values.get("user"),
                global_replicas=values.get("global_replicas"),
                global_concurrency=values.get("global_concurrency"),
                scheduler=values.get("scheduler"),
                compiler=values.get("compiler"),
                worker=values.get("worker"),
                beat=values.get("beat"),
                clean_hook=values.get("clean_hook"),
                api_hooks=values.get("api_hooks"),
                flower=values.get("flower"),
                postgresql=values.get("postgresql"),
                redis=values.get("redis"),
                rabbitmq=values.get("rabbitmq"),
                broker=values.get("broker"),
                email=values.get("email"),
                ldap=values.get("ldap"),
                intervals=values.get("intervals"),
                metrics=values.get("metrics"),
                organization_key=values.get("organization_key"),
                ui=values.get("ui"),
            )

        return values

    @validator("affinity", "celery_affinity", always=True, pre=True)
    def validate_affinity(cls, v):
        return k8s_validation.validate_k8s_affinity(v)

    @validator("tolerations", "celery_tolerations", always=True, pre=True)
    def validate_tolerations(cls, v):
        if not v:
            return v
        return [k8s_validation.validate_k8s_toleration(vi) for vi in v]



from typing import Any, Dict, List, Optional

from pydantic import Field, StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1SectionSpec(BaseResponseModel):
    name: Optional[StrictStr]
    is_minimized: Optional[bool]
    columns: Optional[int]
    height: Optional[int]
    widgets: Optional[List[Dict[str, Any]]]
    page_index: Optional[int] = Field(alias="pageIndex")
    page_size: Optional[int] = Field(alias="pageSize")



from typing import Dict, List, Optional, Union

from clipped.types.ref_or_obj import RefField
from pydantic import Field, StrictStr

from polyaxon.polyflow.params import V1Param
from polyaxon.schemas.base import BaseSchemaModel


class V1Tuner(BaseSchemaModel):
    """You can configure Polyaxon to use a custom tuner to customize the built-in optimizers.

    The tuner allows you to customize the behavior of the operations that
    generate new suggestions based on the previous observations.

    You can provide a queue or provide presets to override
    the default configuration of the component.
    You can resolve any context information from the main operation inside a tuner,
    like params, globals, ...

    To override the complete behavior users can provide their own component.

    Args:
        hub_ref: str
        queue: List[str], optional
        presets: List[str], optional
        params: Dict[str, [V1Param](/docs/core/specification/params/)], optional

    ## YAML usage

    ```yaml
    >>> tuner:
    >>>   hubRef: acme/custom-tuner
    ```

    ## Python usage

    ```python
    >>> from polyaxon.lifecycle import V1Statuses
    >>> from polyaxon.polyflow import V1Tuner
    >>> tuner = V1Tuner(
    >>>     hub_ref="acme/custom-tuner",
    >>>     queue="agent-name/queue-name",
    >>>     persets=["preset1", "preset2"],
    >>> )
    ```

    ## Fields

    ### hubRef

    For several algorithms, Polyaxon provides built-in tuners. these tuners
    are hosted on the public component hub. Users can customize or
    build different service to generate new suggestions.

    To provide a custom component hosted on Polyaxon Component Hub, you can use `hubRef`

    ```yaml
    >>> tuner:
    >>>   hubRef: acme/optimizer-logic:v1
    ...
    ```

    ### queue

    The [queue](/docs/core/scheduling-strategies/queues/) to use.
    If not provided, the default queue will be used.

    ```yaml
    >>> tuner:
    >>>   queue: agent-name/queue-name
    ```

    If the agent name is not specified, Polyaxon will resolve the name of the queue
    based on the default agent.

    ```yaml
    >>> hook:
    >>>   queue: queue-name
    ```

    ### presets

    The [presets](/docs/management/organizations/presets/) to use for the tuner operation,
    if provided, it will override the component's presets otherwise
    the presets of the component will be used if available.

    ```yaml
    >>> tuner:
    >>>   presets: [test]
    ```

    ### params

    The [params](/docs/core/specification/params/) to pass if the handler requires extra params,
    they will be validated against the inputs/outputs.
    If a parameter is passed and the component does not define a corresponding inputs/outputs,
    a validation error will be raised unless the param has the `contextOnly` flag enabled.

    ```yaml
    >>> tuner:
    >>>   params:
    >>>     param1: {value: 1.1}
    >>>     param2: {value: test}
    >>>   ...
    ```
    """

    _IDENTIFIER = "tuner"

    hub_ref: StrictStr = Field(alias="hubRef")
    queue: Optional[StrictStr]
    presets: Optional[Union[List[StrictStr], RefField]]
    params: Optional[Union[Dict[str, V1Param], RefField]]



from typing import Type

from polyaxon.polyaxonfile.specs import kinds
from polyaxon.polyaxonfile.specs.base import BaseSpecification
from polyaxon.polyflow import V1Component


class ComponentSpecification(BaseSpecification):
    """The polyaxonfile specification for component."""

    _SPEC_KIND = kinds.COMPONENT

    CONFIG: Type[V1Component] = V1Component



from typing import Type

from clipped.formatting import Printer

from polyaxon.config.manager import ConfigManager
from polyaxon.schemas.responses.v1_run import V1Run


class RunConfigManager(ConfigManager):
    """Manages run configuration .run file."""

    VISIBILITY = ConfigManager.Visibility.ALL
    IN_PROJECT_DIR = True
    CONFIG_FILE_NAME = ".run"
    CONFIG: Type[V1Run] = V1Run

    @classmethod
    def get_config_or_raise(cls) -> V1Run:
        run = cls.get_config()
        if not run:
            Printer.error("No run was provided.", sys_exit=True)

        return run



import os

from typing import Dict, Type

from polyaxon.config.manager import ConfigManager
from polyaxon.config.spec import ConfigSpec
from polyaxon.schemas.api.home import HomeConfig


class HomeConfigManager(ConfigManager):
    """Manages home configuration .home file."""

    VISIBILITY = ConfigManager.Visibility.GLOBAL
    CONFIG_FILE_NAME = ".home"
    CONFIG: Type[HomeConfig] = HomeConfig

    @classmethod
    def get_config_defaults(cls) -> Dict[str, str]:
        return {"path": cls.get_global_config_path()}

    @classmethod
    def get_config_from_env(cls) -> HomeConfig:
        glob_path = cls.get_global_config_path()
        home_config = cls._CONFIG_READER.read_configs(
            [
                ConfigSpec(glob_path, config_type=".json", check_if_exists=False),
                os.environ,
                {"dummy": "dummy"},
            ]
        )
        return HomeConfig.from_dict(home_config.data)



from polyaxon.compiler.contexts.kubeflow.mpi_job import MPIJobContextsManager
from polyaxon.compiler.contexts.kubeflow.mx_job import MXJobContextsManager
from polyaxon.compiler.contexts.kubeflow.paddle_job import PaddleJobContextsManager
from polyaxon.compiler.contexts.kubeflow.pytroch_job import PytorchJobContextsManager
from polyaxon.compiler.contexts.kubeflow.tf_job import TfJobContextsManager
from polyaxon.compiler.contexts.kubeflow.xgb_job import XGBoostJobContextsManager



from typing import List, Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel
from traceml.artifacts import V1RunArtifact


class V1ListRunArtifactsResponse(BaseResponseModel):
    count: Optional[int]
    results: Optional[List[V1RunArtifact]]
    previous: Optional[StrictStr]
    next: Optional[StrictStr]



from typing import Optional

from pydantic import Extra, Field, StrictStr

from polyaxon.env_vars.keys import EV_KEYS_HOME
from polyaxon.schemas.base import BaseSchemaModel


class HomeConfig(BaseSchemaModel):
    """
    Home config for managing Polyaxon's main context path.


    Args:
        path: `str`. The context path where to write/read configs.
    """

    _IDENTIFIER = "home"

    path: Optional[StrictStr] = Field(alias=EV_KEYS_HOME)

    class Config:
        extra = Extra.ignore



from typing import List, Optional, Union

from clipped.types.ref_or_obj import RefField
from pydantic import StrictStr

from polyaxon.schemas.types.base import BaseTypeConfig


class V1GitType(BaseTypeConfig):
    """Git type allows you to pass a git repo as a parameter.

    If used as an input type, Polyaxon can resolve several git connections
    and will turn this input type into an initializer with logic to clone
    the repo with support for branches and commits,
    the requested repo will be exposed as a context for your jobs and operations.

    Args:
        url: str, optional.
        revision: str, optional.
        flags: List[str], optional

    ### YAML usage

    ### Usage in IO and params definition

    The inputs definition

    ```yaml
    >>> inputs:
    >>>   - name: test1
    >>>     type: git
    >>>   - name: test2
    >>>     type: git
    ```

    The params usage

    ```yaml
    >>> params:
    >>>   test1: {value: {"url": "https://github.com/tensorflow/models"}}
    >>>   test2: {value: {revision: "branchA"}, connection: "my-git-connection"}
    >>>   test3: {
    >>>     value: {flags: ["--recursive", "-c http.sslVerify=false"]},
    >>>     connection: "my-git-connection",
    >>>   }
    ```

    ### Usage in initializers

    ```yaml
    >>> version:  1.1
    >>> kind: component
    >>> run:
    >>>   kind: job
    >>>   init:
    >>>   - git: {"url": "https://github.com/tensorflow/models"}
    >>>   - git:
    >>>       revision: branchA
    >>>     connection: my-git-connection
    >>>   - git:
    >>>       flags: ["--recursive", "-c http.sslVerify=false"]
    >>>     connection: my-git-connection
    >>>   ...
    ```

    ### Python usage

    ### Usage in IO and params definition

    The inputs definition

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.polyflow import V1IO
    >>> inputs = [
    >>>     V1IO(
    >>>         name="test1",
    >>>         type=types.GIT,
    >>>     ),
    >>>     V1IO(
    >>>         name="test2",
    >>>         type=types.GIT,
    >>>     ),
    >>> ]
    ```

    The params usage

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.polyflow import V1Param
    >>> params = {
    >>>     "test1": V1Param(
    >>>         value=types.V1GitType(url="https://github.com/tensorflow/models")
    >>>     ),
    >>>     "test2": V1Param(
    >>>         value=types.V1GitType(revision="branchA"),
    >>>         connection="my-git-connection",
    >>>     ),
    >>> }
    ```

    ### Usage in initializers
    ```python
    >>> from polyaxon.polyflow import V1Component, V1Init, V1Job
    >>> from polyaxon.schemas.types import V1GitType
    >>> from polyaxon.k8s import k8s_schemas
    >>> component = V1Component(
    >>>     run=V1Job(
    >>>        init=[
    >>>             V1Init(
    >>>               git=V1GitType(url="https://github.com/tensorflow/models"),
    >>>             ),
    >>>             V1Init(
    >>>               git=V1GitType(revision="branchA"),
    >>>               connection="my-git-connection",
    >>>             ),
    >>>        ],
    >>>        container=k8s_schemas.V1Container(...)
    >>>     )
    >>> )
    ```
    """

    _IDENTIFIER = "git"

    url: Optional[StrictStr]
    revision: Optional[StrictStr]
    flags: Optional[Union[List[StrictStr], RefField]]

    def get_name(self):
        if self.url:
            return self.url.split("/")[-1].split(".")[0]
        return None



from typing import Dict, List, Optional

from polyaxon.k8s.converter.pod.spec import get_pod_spec, get_pod_template_spec
from polyaxon.k8s.custom_resources.operation import get_operation_custom_object
from polyaxon.k8s.custom_resources.setter import (
    set_collect_logs,
    set_notify,
    set_sync_statuses,
    set_termination,
)
from polyaxon.k8s.replica import ReplicaSpec
from polyaxon.polyflow import V1Notification, V1Termination


def get_spark_replicas_template(
    namespace: str,
    resource_name: str,
    replica_name: str,
    replica: Optional[ReplicaSpec],
    labels: Dict[str, str],
    annotations: Dict[str, str],
    template_spec: Dict,
):
    if not replica:
        return

    metadata, pod_spec = get_pod_spec(
        namespace=namespace,
        main_container=replica.main_container,
        sidecar_containers=replica.sidecar_containers,
        init_containers=replica.init_containers,
        resource_name=resource_name,
        volumes=replica.volumes,
        environment=replica.environment,
        labels=labels,
        annotations=annotations,
    )

    template_spec[replica_name] = {
        "replicas": replica.num_replicas,
        "restartPolicy": pod_spec.restart_policy or "Never",
        "template": get_pod_template_spec(metadata=metadata, pod_spec=pod_spec),
    }


def get_spark_job_custom_resource(
    resource_name: str,
    namespace: str,
    executor: Optional[ReplicaSpec],
    driver: Optional[ReplicaSpec],
    termination: V1Termination,
    collect_logs: bool,
    sync_statuses: bool,
    notifications: List[V1Notification],
    labels: Dict[str, str],
    annotations: Dict[str, str],
) -> Dict:
    template_spec = {}

    get_spark_replicas_template(
        replica_name="Executor",
        replica=executor,
        namespace=namespace,
        resource_name=resource_name,
        labels=labels,
        annotations=annotations,
        template_spec=template_spec,
    )
    get_spark_replicas_template(
        replica_name="Driver",
        replica=driver,
        namespace=namespace,
        resource_name=resource_name,
        labels=labels,
        annotations=annotations,
        template_spec=template_spec,
    )
    template_spec = {"replicaSpecs": template_spec}
    custom_object = {"sparkJobSpec": template_spec}
    custom_object = set_termination(
        custom_object=custom_object, termination=termination
    )
    custom_object = set_collect_logs(
        custom_object=custom_object, collect_logs=collect_logs
    )
    custom_object = set_sync_statuses(
        custom_object=custom_object, sync_statuses=sync_statuses
    )
    custom_object = set_notify(custom_object=custom_object, notifications=notifications)

    return get_operation_custom_object(
        namespace=namespace,
        resource_name=resource_name,
        labels=labels,
        annotations=annotations,
        custom_object=custom_object,
    )



from typing import Any, Dict, List, Optional

from pydantic import StrictStr

from polyaxon.lifecycle import ManagedBy
from polyaxon.schemas import V1RunPending
from polyaxon.schemas.base import BaseResponseModel


class V1OperationBody(BaseResponseModel):
    content: Optional[StrictStr]
    is_managed: Optional[bool]
    managed_by: Optional[ManagedBy]
    pending: Optional[V1RunPending]
    name: Optional[StrictStr]
    description: Optional[StrictStr]
    tags: Optional[List[StrictStr]]
    meta_info: Optional[Dict[str, Any]]



import os

from clipped.utils.http import clean_host

from polyaxon.api import LOCALHOST
from polyaxon.env_vars.keys import EV_KEYS_PLATFORM_HOST


def get_api_host(default: str = LOCALHOST):
    return clean_host(os.environ.get(EV_KEYS_PLATFORM_HOST, default))



from typing import Dict, List, Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1ArtifactTree(BaseResponseModel):
    files: Optional[Dict[str, StrictStr]]
    dirs: Optional[List[StrictStr]]
    is_done: Optional[bool]



import sys

from typing import Any, Dict, Optional

from click import ClickException


class PolyaxonException(ClickException):
    def __init__(self, message=None):
        super().__init__(message)

    def __repr__(self):
        return self.message


class PolyaxonOperatorException(PolyaxonException):
    def __init__(self, cmd, args, return_code, stdout, stderr):
        self.cmd = cmd
        self.args = args
        self.return_code = return_code
        self.stdout = stdout.read() if stdout else None
        self.stderr = stderr.read()
        if stdout:
            message = "`{}` command {} failed with exit status {}\nstdout:\n{}\nstderr:\n{}".format(
                self.cmd, self.args, self.return_code, self.stdout, self.stderr
            )
        else:
            message = "`{}` command {} failed with exit status {}\nstderr:\n{}".format(
                self.cmd, self.args, self.return_code, self.stderr
            )
        super().__init__(message=message)


class PolyaxonConverterError(PolyaxonException):
    pass


class PolyaxonCompilerError(PolyaxonException):
    pass


class PolyTuneException(PolyaxonException):
    pass


class PolyaxonConfigException(PolyaxonException):
    pass


class PolyaxonK8sError(PolyaxonException):
    pass


class PolyaxonAgentError(PolyaxonException):
    pass


class PolyaxonBuildException(PolyaxonException):
    pass


class PolyaxonContainerException(Exception):
    pass


class PolyaxonConnectionError(PolyaxonException):
    pass


class PolyaxonPathException(PolyaxonException):
    pass


class PolyaxonSchemaError(PolyaxonException):
    pass


class PQLException(PolyaxonException):
    pass


class PolyaxonValidationError(PolyaxonSchemaError):
    pass


class PolyaxonfileError(PolyaxonSchemaError):
    pass


class PolyaxonClientException(PolyaxonException):
    pass


class PolyaxonShouldExitError(PolyaxonClientException):
    pass


class PolyaxonHTTPError(PolyaxonClientException):
    def __init__(self, endpoint, response, message=None, status_code=None):
        super().__init__()
        self.endpoint = endpoint
        self.response = response
        self.message = getattr(self, "message", message)
        self.status_code = getattr(self, "status_code", status_code)

    def __str__(self):
        return "{status_code} on {endpoint}.".format(
            status_code=self.status_code, endpoint=self.endpoint
        )


HTTP_ERROR_MESSAGES_MAPPING = {
    400: "Status: 400. One or more request parameters are incorrect",
    401: "Status: 401. Authentication failed. Retry by invoking Polyaxon login.",
    403: "Status: 403. You are not authorized to access this resource on Polyaxon.",
    404: "Status: 404. "
    "The resource you are looking for was not found. Check if the name or uuid is correct.",
    405: "Status: 405. Endpoint does not exist or not configured on this API, "
    "make sure you are connecting to the correct host.",
    429: "Status: 429. You are over the allowed limits for this operation.",
    500: "Status: 502. Internal polyaxon server error, please try again later.",
    502: "Status: 502. Invalid response from Polyaxon server.",
    503: "Status: 503. A problem was encountered, please try again later.",
    504: "Status: 504. Polyaxon server took too long to respond.",
    525: "Status: 525. SSL error.",
}


def handle_api_error(
    e,
    logger: Any,
    message: Optional[str] = None,
    http_messages_mapping: Optional[Dict] = None,
    sys_exit: bool = False,
):
    if http_messages_mapping:
        http_messages_mapping.update(HTTP_ERROR_MESSAGES_MAPPING)
    else:
        http_messages_mapping = HTTP_ERROR_MESSAGES_MAPPING
    if message:
        logger.error(message)
    if e and hasattr(e, "status"):
        if e.status not in http_messages_mapping.keys():
            logger.error("Exception:")
            logger.error(e, stack_info=True, exc_info=True)
        elif getattr(e, "body") and e.status != 404:
            logger.error("Error: %s" % e.body)
        if getattr(e, "reason"):
            logger.error("Reason: %s" % e.reason)
        message = http_messages_mapping.get(e.status)
        if message:
            logger.error(message)
    elif e and hasattr(e, "message"):  # Handling of HTML errors
        error_found = False
        for k in http_messages_mapping.keys():
            if str(k) in e.message:
                logger.error(http_messages_mapping.get(k))
                error_found = True
                break
        if not error_found:
            logger.error("Error:")
            logger.error(e.message)
    elif e:
        logger.error("Exception:")
        logger.error(e, stack_info=True, exc_info=True)
    if sys_exit:
        sys.exit(1)



from typing import Dict, Iterable, List, Optional

from clipped.utils.lists import to_list

from polyaxon.connections import V1Connection, V1ConnectionResource
from polyaxon.k8s import k8s_schemas
from polyaxon.k8s.converter.common.volumes import (
    get_artifacts_context_volume,
    get_configs_context_volume,
    get_connections_context_volume,
    get_docker_context_volume,
    get_shm_context_volume,
    get_volume_from_config_map,
    get_volume_from_connection,
    get_volume_from_secret,
)
from polyaxon.polyflow import V1Init, V1Plugins
from polyaxon.runner.converter.common import constants
from polyaxon.runner.converter.common.volumes import get_volume_name


def get_pod_volumes(
    plugins: Optional[V1Plugins],
    artifacts_store: Optional[V1Connection],
    init_connections: Optional[List[V1Init]],
    connections: List[str],
    connection_by_names: Optional[Dict[str, V1Connection]],
    secrets: Optional[Iterable[V1ConnectionResource]],
    config_maps: Optional[Iterable[V1ConnectionResource]],
    volumes: List[k8s_schemas.V1Volume] = None,
) -> List[k8s_schemas.V1Volume]:
    """Resolve all volumes that need to be mounted"""
    connections = to_list(connections, check_none=True)
    init_connections = to_list(init_connections, check_none=True)
    secrets = to_list(secrets, check_none=True)
    config_maps = to_list(config_maps, check_none=True)
    volumes = to_list(volumes, check_none=True)[:]
    connection_by_names = connection_by_names or {}

    requested_connection_names = connections[:]
    for init_connection in init_connections:
        if (
            init_connection.connection
            and init_connection.connection not in requested_connection_names
        ):
            requested_connection_names.append(init_connection.connection)
    if artifacts_store and artifacts_store.name not in requested_connection_names:
        requested_connection_names.append(artifacts_store.name)

    requested_connections = [connection_by_names[c] for c in requested_connection_names]

    requested_config_maps = V1Connection.get_requested_resources(
        resources=config_maps,
        connections=requested_connections,
        resource_key="config_map",
    )
    requested_secrets = V1Connection.get_requested_resources(
        resources=secrets, connections=requested_connections, resource_key="secret"
    )

    def add_volume_from_connection(connection: V1Connection):
        volume = get_volume_from_connection(connection=connection)
        if volume:
            volumes.append(volume)

    def add_volume_from_resource(resource: V1ConnectionResource, is_secret: bool):
        if is_secret:
            volume = get_volume_from_secret(secret=resource)
        else:
            volume = get_volume_from_config_map(config_map=resource)
        if volume:
            volumes.append(volume)

    volume_names = set()
    connection_ids = set()
    # Handle context volumes from init section
    for init_connection in init_connections:
        volume_name = (
            get_volume_name(init_connection.path)
            if init_connection.path
            else constants.VOLUME_MOUNT_ARTIFACTS
        )
        if volume_name in volume_names:
            continue
        volume_names.add(volume_name)
        volumes.append(get_connections_context_volume(name=volume_name))

    # Add volumes from artifact stores
    for c_name in connection_by_names:
        connection = connection_by_names[c_name]
        if connection.name not in connection_ids:
            connection_ids.add(connection.name)
            add_volume_from_connection(connection=connection)
    # Add volumes from k8s config mount resources
    for secret in requested_secrets:
        add_volume_from_resource(resource=secret, is_secret=True)
    for config_map in requested_config_maps:
        add_volume_from_resource(resource=config_map, is_secret=False)

    # Add logs/outputs stores
    if plugins and (plugins.collect_artifacts or plugins.collect_logs):
        if constants.VOLUME_MOUNT_ARTIFACTS not in volume_names:
            volumes.append(get_artifacts_context_volume())
            volume_names.add(constants.VOLUME_MOUNT_ARTIFACTS)
        if artifacts_store and artifacts_store.name not in connection_ids:
            connection_ids.add(artifacts_store.name)
            add_volume_from_connection(connection=artifacts_store)

    # Add utils contexts
    if plugins and plugins.shm:
        volumes.append(get_shm_context_volume())
    if plugins and plugins.auth:
        volumes.append(get_configs_context_volume())
    if plugins and plugins.docker:
        volumes.append(get_docker_context_volume())
    return volumes



from clipped.utils.lists import to_list
from clipped.utils.strings import strip_spaces


def get_container_command_args(config):
    def sanitize_str(value):
        if not value:
            return
        value = strip_spaces(value=value, join=False)
        value = [c.strip().strip("\\") for c in value if (c and c != "\\")]
        value = [c for c in value if (c and c != "\\")]
        return " ".join(value)

    def sanitize(value):
        return (
            [sanitize_str(v) for v in value]
            if isinstance(value, list)
            else to_list(sanitize_str(value), check_none=True)
        )

    return to_list(config.command, check_none=True), sanitize(config.args)



from typing import Optional

from clipped.types.email import EmailStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1User(BaseResponseModel):
    username: Optional[StrictStr]
    email: Optional[EmailStr]
    name: Optional[StrictStr]
    kind: Optional[StrictStr]
    theme: Optional[int]
    organization: Optional[StrictStr]



from typing import Optional

from clipped.types.email import EmailStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1UserSingup(BaseResponseModel):
    username: Optional[StrictStr]
    email: Optional[EmailStr]
    organization: Optional[StrictStr]
    password: Optional[StrictStr]
    invitation_key: Optional[StrictStr]



from typing import Dict, List, Optional

from clipped.utils.lists import to_list
from pydantic import Field

from polyaxon import settings
from polyaxon.auxiliaries import (
    V1PolyaxonInitContainer,
    V1PolyaxonSidecarContainer,
    get_default_init_container,
    get_default_sidecar_container,
)
from polyaxon.connections import V1Connection, V1ConnectionResource
from polyaxon.exceptions import PolyaxonCompilerError
from polyaxon.polyflow import V1CompiledOperation, V1Init
from polyaxon.schemas.base import BaseSchemaModel
from polyaxon.schemas.cli.agent_config import AgentConfig


class AgentResolver(BaseSchemaModel):
    polyaxon_sidecar: Optional[V1PolyaxonSidecarContainer]
    polyaxon_init: Optional[V1PolyaxonInitContainer]
    namespace: Optional[str]
    secrets: Optional[List[V1ConnectionResource]]
    config_maps: Optional[List[V1ConnectionResource]]
    connection_by_names: Optional[Dict[str, V1Connection]]
    artifacts_store: Optional[V1Connection]
    default_sa: Optional[str]
    internal_auth: Optional[bool] = Field(default=False)

    def resolve(
        self, compiled_operation: V1CompiledOperation, agent_config: AgentConfig = None
    ):
        if not agent_config and settings.AGENT_CONFIG:
            agent_config = settings.AGENT_CONFIG.clone()
        if not agent_config:
            raise PolyaxonCompilerError(
                "Agent configuration not found or agent not configured."
            )

        self.default_sa = agent_config.runs_sa
        self._resolve_run_connections(
            compiled_operation=compiled_operation, agent_config=agent_config
        )
        self.artifacts_store = agent_config.artifacts_store

        self.secrets = agent_config.secrets
        self.config_maps = agent_config.config_maps

        self.polyaxon_sidecar = agent_config.sidecar or get_default_sidecar_container()
        self.polyaxon_init = agent_config.init or get_default_init_container()
        self.namespace = agent_config.namespace
        self.polyaxon_sidecar.monitor_logs = agent_config.is_replica

    def _resolve_run_connections(
        self, compiled_operation: V1CompiledOperation, agent_config: AgentConfig
    ):
        if not self.connection_by_names:
            self.connection_by_names = {}
        if agent_config.artifacts_store:  # Resolve default artifacts store
            self.connection_by_names[
                agent_config.artifacts_store.name
            ] = agent_config.artifacts_store

        if (
            compiled_operation.is_job_run
            or compiled_operation.is_service_run
            or compiled_operation.is_notifier_run
            or compiled_operation.is_tuner_run
        ):
            self._resolve_replica_connections(
                init=compiled_operation.run.init,
                connections=compiled_operation.run.connections,
                agent_config=agent_config,
            )
        if compiled_operation.is_distributed_run:
            self._resolve_distributed_connections(
                compiled_operation=compiled_operation, agent_config=agent_config
            )
        if compiled_operation.is_dag_run:
            self._resolve_connections(
                connections=compiled_operation.run.connections,
                agent_config=agent_config,
            )

    def _resolve_connections(self, connections: List[str], agent_config: AgentConfig):
        if connections:
            connection_by_names = self.connection_by_names or {}
            missing_connections = set()
            for c in connections:
                if c not in agent_config.connections_by_names:
                    missing_connections.add(c)
                else:
                    connection_by_names[c] = agent_config.connections_by_names[c]
            if missing_connections:
                raise PolyaxonCompilerError(
                    "Some connection refs were provided "
                    "but were not found in the "
                    "agent.connections catalog: `{}`".format(missing_connections)
                )
            self.connection_by_names = connection_by_names

    def _resolve_replica_connections(
        self, connections: List[str], init: List[V1Init], agent_config: AgentConfig
    ):
        connections = to_list(connections, check_none=True)
        self._resolve_connections(connections=connections, agent_config=agent_config)
        init = to_list(init, check_none=True)
        init = [i.connection for i in init if i.connection]
        self._resolve_connections(connections=init, agent_config=agent_config)

    def _resolve_distributed_connections(
        self, compiled_operation: V1CompiledOperation, agent_config: AgentConfig
    ):
        def _resolve_replica(replica):
            if not replica:
                return
            self._resolve_replica_connections(
                init=replica.init,
                connections=replica.connections,
                agent_config=agent_config,
            )

        if compiled_operation.is_mpi_job_run:
            _resolve_replica(compiled_operation.run.launcher)
            _resolve_replica(compiled_operation.run.worker)
        elif compiled_operation.is_tf_job_run:
            _resolve_replica(compiled_operation.run.chief)
            _resolve_replica(compiled_operation.run.worker)
            _resolve_replica(compiled_operation.run.ps)
            _resolve_replica(compiled_operation.run.evaluator)
        elif (
            compiled_operation.is_pytorch_job_run
            or compiled_operation.is_paddle_job_run
        ):
            _resolve_replica(compiled_operation.run.master)
            _resolve_replica(compiled_operation.run.worker)
        elif compiled_operation.is_mx_job_run:
            _resolve_replica(compiled_operation.run.scheduler)
            _resolve_replica(compiled_operation.run.worker)
            _resolve_replica(compiled_operation.run.server)
            _resolve_replica(compiled_operation.run.tuner)
            _resolve_replica(compiled_operation.run.tuner_tracker)
            _resolve_replica(compiled_operation.run.tuner_server)
        elif compiled_operation.is_xgb_job_run:
            _resolve_replica(compiled_operation.run.master)
            _resolve_replica(compiled_operation.run.worker)



from typing import Dict

from polyaxon.api import (
    EXTERNAL_V1_LOCATION,
    REWRITE_EXTERNAL_V1_LOCATION,
    REWRITE_SERVICES_V1_LOCATION,
    SERVICES_V1_LOCATION,
)
from polyaxon.compiler.contexts.base import BaseContextsManager
from polyaxon.connections import V1Connection
from polyaxon.polyflow import V1CompiledOperation
from polyaxon.utils.urls_utils import get_proxy_run_url


class ServiceContextsManager(BaseContextsManager):
    @classmethod
    def resolve(
        cls,
        namespace: str,
        owner_name: str,
        project_name: str,
        run_uuid: str,
        contexts: Dict,
        compiled_operation: V1CompiledOperation,
        connection_by_names: Dict[str, V1Connection],
    ) -> Dict:
        contexts = cls._resolver_replica(
            contexts=contexts,
            init=compiled_operation.run.init,
            connections=compiled_operation.run.connections,
            connection_by_names=connection_by_names,
        )
        if compiled_operation.is_service_run:
            contexts["globals"]["ports"] = compiled_operation.run.ports
            if compiled_operation.run.is_external:
                service = (
                    REWRITE_EXTERNAL_V1_LOCATION
                    if compiled_operation.run.rewrite_path
                    else EXTERNAL_V1_LOCATION
                )
            else:
                service = (
                    REWRITE_SERVICES_V1_LOCATION
                    if compiled_operation.run.rewrite_path
                    else SERVICES_V1_LOCATION
                )
            base_url = get_proxy_run_url(
                service=service,
                namespace=namespace,
                owner=owner_name,
                project=project_name,
                run_uuid=run_uuid,
            )
            contexts["globals"]["base_url"] = base_url

        return contexts



from typing import Dict, Iterable, Optional

from polyaxon import pkg
from polyaxon.connections import V1Connection, V1ConnectionResource
from polyaxon.k8s.converter.base import BaseConverter
from polyaxon.k8s.converter.mixins import SparkJobMixin
from polyaxon.k8s.custom_resources.spark_job import get_spark_job_custom_resource
from polyaxon.k8s.replica import ReplicaSpec
from polyaxon.polyflow import V1CompiledOperation, V1Plugins, V1SparkJob, V1SparkReplica


class SparkJobConverter(SparkJobMixin, BaseConverter):
    def get_resource(
        self,
        compiled_operation: V1CompiledOperation,
        artifacts_store: V1Connection,
        connection_by_names: Dict[str, V1Connection],
        secrets: Optional[Iterable[V1ConnectionResource]],
        config_maps: Optional[Iterable[V1ConnectionResource]],
        default_sa: Optional[str] = None,
        default_auth: bool = False,
    ) -> Dict:
        job = compiled_operation.run  # type: V1SparkJob

        def _get_replica(replica: Optional[V1SparkReplica]) -> Optional[ReplicaSpec]:
            if not replica:
                return None
            return self.get_replica_resource(
                plugins=plugins,
                environment=replica.environment,
                volumes=job.volumes or [],
                init=replica.init or [],
                sidecars=replica.sidecars or [],
                container=replica.container,
                artifacts_store=artifacts_store,
                connections=job.connections or [],
                connection_by_names=connection_by_names,
                secrets=secrets,
                config_maps=config_maps,
                kv_env_vars=kv_env_vars,
                default_sa=default_sa,
                num_replicas=replica.replicas,
            )

        kv_env_vars = compiled_operation.get_env_io()
        plugins = V1Plugins.get_or_create(
            config=compiled_operation.plugins, auth=default_auth
        )
        executor = _get_replica(job.executor)
        driver = _get_replica(job.driver)
        labels = self.get_labels(version=pkg.VERSION, labels={})

        return get_spark_job_custom_resource(
            namespace=self.namespace,
            resource_name=self.resource_name,
            executor=executor,
            driver=driver,
            termination=compiled_operation.termination,
            collect_logs=plugins.collect_logs,
            sync_statuses=plugins.sync_statuses,
            notifications=plugins.notifications,
            labels=labels,
            annotations=self.annotations,
        )



from typing import Any, Dict, Optional

from clipped.types.email import EmailStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1TrialStart(BaseResponseModel):
    name: Optional[StrictStr]
    email: Optional[EmailStr]
    organization: Optional[StrictStr]
    plan: Optional[StrictStr]
    seats: Optional[int]
    details: Optional[Dict[str, Any]]



import ast
import jinja2

from collections.abc import Mapping
from datetime import date, datetime, timedelta
from enum import Enum
from typing import Dict
from uuid import UUID

from clipped.utils.serialization import (
    date_serialize,
    datetime_serialize,
    timedelta_serialize,
    uuid_serialize,
)

from polyaxon.exceptions import PolyaxonSchemaError
from polyaxon.polyaxonfile.specs.libs.engine import get_engine
from polyaxon.polyaxonfile.specs.sections import Sections
from polyaxon.polyflow import ParamSpec

try:
    import numpy as np
except (ImportError, ModuleNotFoundError):
    np = None


class PolyaxonfileParser:
    """Parses the Polyaxonfile."""

    engine = get_engine()

    @staticmethod
    def _get_section_data(section_data):
        if hasattr(section_data, "to_dict"):
            return section_data.to_dict()
        return section_data

    @classmethod
    def _get_section(cls, config, section):
        if not hasattr(config, section):
            return None
        section_data = getattr(config, section)
        if isinstance(section_data, list):
            return [cls._get_section_data(d) for d in section_data]

        return cls._get_section_data(section_data)

    @staticmethod
    def get_parsed_params(param_spec: Dict[str, ParamSpec]) -> Dict:
        parsed_params = {"params": {}}
        as_arg = []
        for param in param_spec or {}:
            parsed_params[param] = param_spec[param].get_display_value()
            # Add params level
            if param_spec[param].is_requested:
                as_arg.append(param_spec[param].as_arg())
                parsed_params["params"][param] = param_spec[param].to_parsed_param()

        parsed_params["params"]["as_args"] = [a for a in as_arg if a]
        return parsed_params

    @classmethod
    def parse_operation(
        cls, config, param_spec: Dict[str, ParamSpec]
    ):  # pylint:disable=too-many-branches
        parsed_params = cls.get_parsed_params(param_spec)

        parsed_data = {Sections.VERSION: config.version, Sections.KIND: config.kind}

        if config.name:
            parsed_data[Sections.NAME] = config.name
        if config.description:
            parsed_data[Sections.DESCRIPTION] = config.description
        if config.tags:
            parsed_data[Sections.TAGS] = config.tags
        contexts = getattr(config, Sections.CONTEXTS)
        if contexts:
            parsed_data[Sections.CONTEXTS] = [
                cls.parse_io(io.to_dict(), parsed_params) for io in contexts
            ]
        inputs = getattr(config, Sections.INPUTS)
        if inputs:
            parsed_data[Sections.INPUTS] = [
                cls.parse_io(io.to_dict(), parsed_params) for io in inputs
            ]
        outputs = getattr(config, Sections.OUTPUTS)
        if outputs:
            parsed_data[Sections.OUTPUTS] = [
                cls.parse_io(io.to_dict(), parsed_params) for io in outputs
            ]

        # Check workflow
        for section in Sections.PARSING_SECTIONS:
            config_section = cls._get_section(config, section)
            if config_section:
                parsed_data[section] = cls.parse_expression(
                    config_section, parsed_params
                )

        parsed_data = cls._parse_conditions(parsed_data)

        config_section = cls._get_section(config, Sections.RUN)
        if config_section:
            parsed_data[Sections.RUN] = config_section
        config_section = cls._get_section(config, Sections.BUILD)
        if config_section:
            parsed_data[Sections.BUILD] = config_section
        config_section = cls._get_section(config, Sections.HOOKS)
        if config_section:
            parsed_data[Sections.HOOKS] = config_section

        return parsed_data

    @classmethod
    def _parse_conditions(cls, parsed_data: Dict) -> Dict:
        if Sections.CONDITIONS in parsed_data and not isinstance(
            parsed_data[Sections.CONDITIONS], str
        ):
            parsed_data[Sections.CONDITIONS] = str(parsed_data[Sections.CONDITIONS])
        if Sections.SKIP_ON_UPSTREAM_SKIP in parsed_data and not isinstance(
            parsed_data[Sections.SKIP_ON_UPSTREAM_SKIP], str
        ):
            parsed_data[Sections.SKIP_ON_UPSTREAM_SKIP] = str(
                parsed_data[Sections.SKIP_ON_UPSTREAM_SKIP]
            )
        if Sections.IS_APPROVED in parsed_data and not isinstance(
            parsed_data[Sections.IS_APPROVED], str
        ):
            parsed_data[Sections.IS_APPROVED] = str(parsed_data[Sections.IS_APPROVED])
        return parsed_data

    @classmethod
    def _parse_hook(cls, config, param_spec: Dict[str, ParamSpec]):
        hook = cls.parse_section(config, param_spec=param_spec, parse_params=True)
        return cls._parse_conditions(hook)

    @classmethod
    def parse_hooks(cls, config, param_spec: Dict[str, ParamSpec]):
        hooks = config.hooks or []
        return [
            cls._parse_hook(hook.to_dict(), param_spec=param_spec) for hook in hooks
        ]

    @classmethod
    def parse_runtime(cls, parsed_data, param_spec: Dict[str, ParamSpec]):
        config_section = cls.parse_section(
            parsed_data.get(Sections.RUN), param_spec=param_spec, parse_params=True
        )
        if config_section:
            parsed_data[Sections.RUN] = config_section
        return parsed_data

    @classmethod
    def parse_distributed_runtime(
        cls, parsed_data, param_spec: Dict[str, Dict[str, ParamSpec]]
    ):
        run_section = parsed_data.get(Sections.RUN)
        config_section = {}
        for k in run_section:
            if k in param_spec:
                config_section[k] = cls.parse_section(
                    run_section[k], param_spec=param_spec[k], parse_params=True
                )
            else:
                config_section[k] = run_section[k]
        if config_section:
            parsed_data[Sections.RUN] = config_section
        return parsed_data

    @classmethod
    def parse_section(
        cls, config_section, param_spec: Dict[str, ParamSpec], parse_params: bool = True
    ):
        if parse_params:
            param_spec = cls.get_parsed_params(param_spec)
        if config_section:
            return cls.parse_expression(config_section, param_spec)
        return config_section

    @classmethod
    def parse_io(
        cls,
        expression: Dict,
        params: Dict[str, ParamSpec],
        check_operators: bool = False,
    ):
        # Arg format should not be converted
        arg_format = expression.pop("argFormat", None)
        value = cls.parse_expression(expression, params, check_operators)
        if arg_format:
            value["argFormat"] = arg_format
        return value

    @classmethod
    def parse_expression(  # pylint:disable=too-many-branches
        cls, expression, params: Dict, check_operators: bool = False
    ):
        try:
            return cls._parse_expression(expression, params, check_operators)
        except jinja2.exceptions.TemplateError as e:
            raise PolyaxonSchemaError(
                "Encountered a problem parsing the template, "
                "please make sure your variables are resolvable. "
                "Error: {}".format(repr(e))
            )

    @classmethod
    def _parse_expression(  # pylint:disable=too-many-branches
        cls, expression, params: Dict, check_operators: bool = False
    ):
        if isinstance(expression, (int, float, complex, type(None))):
            return expression
        if isinstance(expression, Enum):
            return expression.value
        if isinstance(expression, datetime):
            return datetime_serialize(expression)
        if isinstance(expression, date):
            return date_serialize(expression)
        if isinstance(expression, timedelta):
            return timedelta_serialize(expression)
        if isinstance(expression, UUID):
            return uuid_serialize(expression)
        if np and isinstance(expression, np.integer):
            return int(expression)
        if np and isinstance(expression, np.floating):
            return float(expression)
        if isinstance(expression, Mapping):
            if len(expression) == 1:
                old_key, value = list(expression.items())[0]
                # always parse the keys, they must be base object or evaluate to base objects
                key = cls._parse_expression(old_key, params)
                if check_operators and cls.is_operator(key):
                    return cls._parse_operator({key: value}, params)
                else:
                    return {key: cls.parse_expression(value, params, check_operators)}

            new_expression = {}
            for k, v in expression.items():
                new_expression.update(
                    cls._parse_expression({k: v}, params, check_operators)
                )
            return new_expression

        if isinstance(expression, list):
            return list(
                cls.parse_expression(v, params, check_operators) for v in expression
            )
        if isinstance(expression, tuple):
            return tuple(
                cls.parse_expression(v, params, check_operators) for v in expression
            )
        if isinstance(expression, str):
            return cls._evaluate_expression(expression, params, check_operators)

    @classmethod
    def _evaluate_expression(cls, expression, params, check_operators):
        try:
            result = cls.engine.from_string(expression).render(**params)
        except (ValueError, TypeError) as e:
            raise PolyaxonSchemaError(
                "Encountered a problem parsing the template, "
                "please make sure your variables are resolvable. "
                "Error: {}".format(repr(e))
            )
        if result == expression:
            try:
                return ast.literal_eval(result)
            except (ValueError, SyntaxError):
                pass
            return result
        return cls.parse_expression(result, params, check_operators)

    @classmethod
    def _parse_operator(cls, expression, params):
        k, v = list(expression.items())[0]
        op = Sections.OPERATORS[k].from_dict(v)
        return op.parse(parser=cls, params=params)

    @staticmethod
    def is_operator(key):
        return key in Sections.OPERATORS



# Platform
EV_KEYS_PLATFORM_CONFIG = "POLYAXON_PLATFORM_CONFIG"

# Context path for Polyaxon's config files
EV_KEYS_HOME = "POLYAXON_HOME"

# General
EV_KEYS_NO_CONFIG = "POLYAXON_NO_CONFIG"
EV_KEYS_SERVICE = "POLYAXON_SERVICE"
EV_KEYS_ENVIRONMENT = "POLYAXON_ENVIRONMENT"
EV_KEYS_HEADER = "POLYAXON_HEADER"
EV_KEYS_HEADER_SERVICE = "POLYAXON_HEADER_SERVICE"
EV_KEYS_AUTHENTICATION_TYPE = "POLYAXON_AUTHENTICATION_TYPE"
EV_KEYS_DEBUG = "POLYAXON_DEBUG"
EV_KEYS_TIMEOUT = "POLYAXON_TIMEOUT"
EV_KEYS_TRACKING_TIMEOUT = "POLYAXON_TRACKING_TIMEOUT"
EV_KEYS_TIME_ZONE = "POLYAXON_TIME_ZONE"
EV_KEYS_WATCH_INTERVAL = "POLYAXON_WATCH_INTERVAL"
EV_KEYS_INTERVAL = "POLYAXON_INTERVAL"
EV_KEYS_LOG_LEVEL = "POLYAXON_LOG_LEVEL"
EV_KEYS_K8S_NAMESPACE = "POLYAXON_K8S_NAMESPACE"
EV_KEYS_K8S_NODE_NAME = "POLYAXON_K8S_NODE_NAME"
EV_KEYS_K8S_POD_ID = "POLYAXON_K8S_POD_ID"
EV_KEYS_PLATFORM_HOST = "POLYAXON_PLATFORM_HOST"
EV_KEYS_HOST = "POLYAXON_HOST"
EV_KEYS_API_VERSION = "POLYAXON_API_VERSION"
EV_KEYS_VERIFY_SSL = "POLYAXON_VERIFY_SSL"
EV_KEYS_SSL_CA_CERT = "POLYAXON_SSL_CA_CERT"
EV_KEYS_CERT_FILE = "POLYAXON_CERT_FILE"
EV_KEYS_KEY_FILE = "POLYAXON_KEY_FILE"
EV_KEYS_ASSERT_HOSTNAME = "POLYAXON_ASSERT_HOSTNAME"
EV_KEYS_CONNECTION_POOL_MAXSIZE = "POLYAXON_CONNECTION_POOL_MAXSIZE"
EV_KEYS_LOGS_ROOT = "POLYAXON_LOG_ROOT"
EV_KEYS_ARCHIVES_ROOT = "POLYAXON_ARCHIVES_ROOT"
EV_KEYS_ARTIFACTS_ROOT = "POLYAXON_ARTIFACTS_ROOT"
EV_KEYS_STATIC_ROOT = "POLYAXON_STATIC_ROOT"
EV_KEYS_CONTEXT_ROOT = "POLYAXON_CONTEXT_ROOT"
EV_KEYS_OFFLINE_ROOT = "POLYAXON_OFFLINE_ROOT"
EV_KEYS_DISABLE_ERRORS_REPORTING = "POLYAXON_DISABLE_ERRORS_REPORTING"
EV_KEYS_INTERVALS_COMPATIBILITY_CHECK = "POLYAXON_INTERVALS_COMPATIBILITY_CHECK"
EV_KEYS_RETRIES = "POLYAXON_RETRIES"
EV_KEYS_UPLOAD_SIZE = "POLYAXON_UPLOAD_SIZE"
EV_KEYS_MAX_CONCURRENCY = "POLYAXON_MAX_CONCURRENCY"
EV_KEYS_HAS_PROCESS_SIDECAR = "POLYAXON_HAS_PROCESS_SIDECAR"

# Secrets
EV_KEYS_SECRET_KEY = "POLYAXON_SECRET_KEY"  # noqa
EV_KEYS_SECRET_INTERNAL_TOKEN = "POLYAXON_SECRET_INTERNAL_TOKEN"  # noqa
EV_KEYS_AUTH_TOKEN = "POLYAXON_AUTH_TOKEN"
EV_KEYS_AUTH_USERNAME = "POLYAXON_AUTH_USERNAME"
EV_KEYS_CUSTOM_ERRORS_OPTIONS = "POLYAXON_CUSTOM_ERRORS_OPTIONS"

# Containers
EV_KEYS_AGENT_INSTANCE = "POLYAXON_AGENT_INSTANCE"
EV_KEYS_RUN_INSTANCE = "POLYAXON_RUN_INSTANCE"
EV_KEYS_CONTAINER_ID = "POLYAXON_CONTAINER_ID"
EV_KEYS_HASH_LENGTH = "POLYAXON_HASH_LENGTH"

# Flags
EV_KEYS_K8S_IN_CLUSTER = "POLYAXON_K8S_IN_CLUSTER"
EV_KEYS_IS_MANAGED = "POLYAXON_IS_MANAGED"
EV_KEYS_IS_OFFLINE = "POLYAXON_IS_OFFLINE"
EV_KEYS_NO_OP = "POLYAXON_NO_OP"
EV_KEYS_NO_API = "POLYAXON_NO_API"

# Registry
EV_KEYS_USE_GIT_REGISTRY = "POLYAXON_USE_GIT_REGISTRY"
EV_KEYS_PUBLIC_REGISTRY = "POLYAXON_PUBLIC_REGISTRY"

# Agent
EV_KEYS_AGENT_SIDECAR = "POLYAXON_AGENT_SIDECAR"
EV_KEYS_AGENT_INIT = "POLYAXON_AGENT_INIT"
EV_KEYS_AGENT_CLEANER = "POLYAXON_AGENT_CLEANER"
EV_KEYS_AGENT_NOTIFIER = "POLYAXON_AGENT_NOTIFIER"
EV_KEYS_AGENT_DEFAULT_SCHEDULING = "POLYAXON_AGENT_DEFAULT_SCHEDULING"
EV_KEYS_AGENT_DEFAULT_IMAGE_PULL_SECRETS = "POLYAXON_AGENT_DEFAULT_IMAGE_PULL_SECRETS"
EV_KEYS_AGENT_ARTIFACTS_STORE = "POLYAXON_AGENT_ARTIFACTS_STORE"
EV_KEYS_AGENT_CONNECTIONS = "POLYAXON_AGENT_CONNECTIONS"
EV_KEYS_SET_AGENT = "POLYAXON_SET_AGENT"
EV_KEYS_K8S_APP_SECRET_NAME = "POLYAXON_K8S_APP_SECRET_NAME"  # noqa
EV_KEYS_AGENT_SECRET_NAME = "POLYAXON_AGENT_SECRET_NAME"  # noqa
EV_KEYS_AGENT_RUNS_SA = "POLYAXON_AGENT_RUNS_SA"
EV_KEYS_AGENT_EXECUTOR_REFRESH_INTERVAL = "POLYAXON_AGENT_EXECUTOR_REFRESH_INTERVAL"
EV_KEYS_AGENT_IS_REPLICA = "POLYAXON_AGENT_IS_REPLICA"
EV_KEYS_AGENT_USE_PROXY_ENV_VARS_IN_OPS = "POLYAXON_AGENT_USE_PROXY_ENV_VARS_IN_OPS"

# Connections
EV_KEYS_COLLECT_ARTIFACTS = "POLYAXON_COLLECT_ARTIFACTS"
EV_KEYS_COLLECT_RESOURCES = "POLYAXON_COLLECT_RESOURCES"
EV_KEYS_ARTIFACTS_STORE_NAME = "POLYAXON_ARTIFACTS_STORE_NAME"
EV_KEYS_GIT_CREDENTIALS = "POLYAXON_GIT_CREDENTIALS"
EV_KEYS_GIT_CREDENTIALS_STORE = "POLYAXON_GIT_CREDENTIALS_STORE"
EV_KEYS_SSH_PATH = "POLYAXON_SSH_PATH"
EV_KEYS_SSH_PRIVATE_KEY = "POLYAXON_SSH_PRIVATE_KEY"

# Ops
EV_KEYS_ISTIO_ENABLED = "POLYAXON_ISTIO_ENABLED"
EV_KEYS_TFJOB_ENABLED = "POLYAXON_TFJOB_ENABLED"
EV_KEYS_PYTORCH_JOB_ENABLED = "POLYAXON_PYTORCH_JOB_ENABLED"
EV_KEYS_PADDLE_JOB_ENABLED = "POLYAXON_PADDLE_JOB_ENABLED"
EV_KEYS_MXJOB_ENABLED = "POLYAXON_MXJOB_ENABLED"
EV_KEYS_XGBOOST_JOB_ENABLED = "POLYAXON_XGBOOST_JOB_ENABLED"
EV_KEYS_MPIJOB_ENABLED = "POLYAXON_MPIJOB_ENABLED"
EV_KEYS_RAYJOB_ENABLED = "POLYAXON_RAYJOB_ENABLED"
EV_KEYS_SPARKJOB_ENABLED = "POLYAXON_SPARKJOB_ENABLED"
EV_KEYS_DASKJOB_ENABLED = "POLYAXON_DASKJOB_ENABLED"

# Sandbox
EV_KEYS_SANDBOX_PORT = "POLYAXON_SANDBOX_PORT"
EV_KEYS_SANDBOX_HOST = "POLYAXON_SANDBOX_HOST"
EV_KEYS_SANDBOX_DEBUG = "POLYAXON_SANDBOX_DEBUG"
EV_KEYS_SANDBOX_SSL_ENABLED = "POLYAXON_SANDBOX_SSL_ENABLED"
EV_KEYS_SANDBOX_WORKERS = "POLYAXON_SANDBOX_WORKERS"
EV_KEYS_SANDBOX_PER_CORE = "POLYAXON_SANDBOX_PER_CORE"
EV_KEYS_SANDBOX_IS_LOCAL = "POLYAXON_SANDBOX_IS_LOCAL"
EV_KEYS_SANDBOX_ROOT = "POLYAXON_SANDBOX_ROOT"

# Proxies
EV_KEYS_PROXY_NAMESPACES = "POLYAXON_PROXY_NAMESPACES"
EV_KEYS_PROXY_GATEWAY_PORT = "POLYAXON_PROXY_GATEWAY_PORT"
EV_KEYS_PROXY_GATEWAY_TARGET_PORT = "POLYAXON_PROXY_GATEWAY_TARGET_PORT"
EV_KEYS_PROXY_GATEWAY_HOST = "POLYAXON_PROXY_GATEWAY_HOST"
EV_KEYS_PROXY_STREAMS_PORT = "POLYAXON_PROXY_STREAMS_PORT"
EV_KEYS_PROXY_STREAMS_TARGET_PORT = "POLYAXON_PROXY_STREAMS_TARGET_PORT"
EV_KEYS_PROXY_STREAMS_HOST = "POLYAXON_PROXY_STREAMS_HOST"
EV_KEYS_PROXY_API_PORT = "POLYAXON_PROXY_API_PORT"
EV_KEYS_PROXY_API_TARGET_PORT = "POLYAXON_PROXY_API_TARGET_PORT"
EV_KEYS_PROXY_API_HOST = "POLYAXON_PROXY_API_HOST"
EV_KEYS_PROXY_LOCAL_PORT = "POLYAXON_PROXY_LOCAL_PORT"
EV_KEYS_PROXY_API_USE_RESOLVER = "POLYAXON_PROXY_API_USE_RESOLVER"
EV_KEYS_PROXY_SERVICES_PORT = "POLYAXON_PROXY_SERVICES_PORT"
EV_KEYS_PROXY_SSL_PATH = "POLYAXON_PROXY_SSL_PATH"
EV_KEYS_PROXY_SSL_ENABLED = "POLYAXON_PROXY_SSL_ENABLED"
EV_KEYS_PROXY_AUTH_ENABLED = "POLYAXON_PROXY_AUTH_ENABLED"
EV_KEYS_PROXY_AUTH_EXTERNAL = "POLYAXON_PROXY_AUTH_EXTERNAL"
EV_KEYS_PROXY_AUTH_USE_RESOLVER = "POLYAXON_PROXY_AUTH_USE_RESOLVER"
EV_KEYS_PROXY_HAS_FORWARD_PROXY = "POLYAXON_PROXY_HAS_FORWARD_PROXY"
EV_KEYS_PROXY_FORWARD_PROXY_PORT = "POLYAXON_PROXY_FORWARD_PROXY_PORT"
EV_KEYS_PROXY_FORWARD_PROXY_HOST = "POLYAXON_PROXY_FORWARD_PROXY_HOST"
EV_KEYS_PROXY_FORWARD_PROXY_KIND = "POLYAXON_PROXY_FORWARD_PROXY_KIND"
EV_KEYS_UI_IN_SANDBOX = "POLYAXON_UI_IN_SANDBOX"
EV_KEYS_UI_ADMIN_ENABLED = "POLYAXON_UI_ADMIN_ENABLED"
EV_KEYS_UI_ASSETS_VERSION = "POLYAXON_UI_ASSETS_VERSION"
EV_KEYS_UI_ENABLED = "POLYAXON_UI_ENABLED"
EV_KEYS_UI_OFFLINE = "POLYAXON_UI_OFFLINE"
EV_KEYS_UI_BASE_URL = "POLYAXON_UI_BASE_URL"
EV_KEYS_STATIC_URL = "POLYAXON_STATIC_URL"
EV_KEYS_DNS_USE_RESOLVER = "POLYAXON_DNS_USE_RESOLVER"
EV_KEYS_DNS_CUSTOM_CLUSTER = "POLYAXON_DNS_CUSTOM_CLUSTER"
EV_KEYS_DNS_BACKEND = "POLYAXON_DNS_BACKEND"
EV_KEYS_DNS_PREFIX = "POLYAXON_DNS_PREFIX"
EV_KEYS_NGINX_TIMEOUT = "POLYAXON_NGINX_TIMEOUT"
EV_KEYS_NGINX_INDENT_CHAR = "POLYAXON_NGINX_INDENT_CHAR"
EV_KEYS_NGINX_INDENT_WIDTH = "POLYAXON_NGINX_INDENT_WIDTH"
